import re
from collections import OrderedDict

from clldutils.misc import lazyproperty


patrons = {
    'Alena': 'AWM',
    'Alena Witzlack-Makarevich': 'AWM',
    'Hannah': 'HJH',
    'Hannah J. Haynie': 'HJH',
    'Harald': 'HH',
    'Harald Hammarström': 'HH',
    'Hedvig': 'HS',
    'Hedvig Skirgård': 'HS',
    'Jakob': 'JLE',
    'Jakob Lesage': 'JLE',
    'Jeremy': 'JC',
    'Jeremy Collins': 'JC',
    'Jay': 'JLA',
    'Jay Latarche': 'JLA',
}


class Feature(OrderedDict):
    def __init__(self, spec, wiki):
        OrderedDict.__init__(self, spec)
        self._wiki = wiki
        self.domain = OrderedDict()
        spec = self['Possible Values'].replace('multistate', '').strip()
        delimiter = ',' if ',' in spec else ';'
        for val in spec.split(delimiter):
            val, desc = re.split(r'\s*:\s*', val.strip())
            int(val)
            self.domain[val] = desc

    @classmethod
    def from_chunk(cls, chunk, wiki):
        items = []
        for line in chunk.split('\n'):
            if ':' in line and not line.startswith('#'):
                items.append(re.split(r'\s*:\s*', line.strip(), 1))
        # Make sure there are no duplicate keys:
        assert len(set(i[0] for i in items)) == len(items)
        res = cls(items, wiki)
        for k, v in res.wiki.items():
            if k in res:
                if res[k] != v:  # pragma: no cover
                    print('++++', k, res[k], v)
        return res

    def as_chunk(self):
        return ''.join('{0}: {1}\n'.format(k, v) for k, v in self.items())

    @lazyproperty
    def wiki(self):
        res = OrderedDict()
        p = self._wiki / '{}.md'.format(self.id)
        if not p.exists():
            return res
        title, header, lines = None, None, []
        for line in p.read_text(encoding='utf-8-sig').split('\n'):
            line = line.strip()
            if (not title) and line.startswith('#'):
                res['title'] = title = line.replace('#', '').strip()
            elif line.startswith('## '):
                if lines:
                    res[header] = '\n'.join(lines).strip()
                    lines = []
                header = line.replace('## ', '').strip()
            else:
                lines.append(line)
        if lines:
            res[header] = '\n'.join(lines).strip()
        res['Patron'] = res['Patron'].split('\n')[0]
        return res

    @lazyproperty
    def description(self):
        if self._wiki.joinpath('{0}.md'.format(self.id)).exists():
            return self._wiki.joinpath('{0}.md'.format(self.id)).read_text(encoding='utf-8-sig')
        return self['Clarifying Comments']  # pragma: no cover

    @lazyproperty
    def id(self):
        return self['Grambank ID']

    @lazyproperty
    def name(self):
        return self['Feature']

    @lazyproperty
    def patrons(self):
        return [
            patrons[k] for k in
            re.split(r'\s+(?:&|and)\s+', self.wiki['Patron'].strip())]


class GB20(object):
    CHUNK_SEP = '\n\n\n\n'

    def __init__(self, path):
        self.path = path

    def iterfeatures(self, wiki):
        for chunk in self.path.read_text(encoding='utf-8').split(self.CHUNK_SEP):
            if chunk.strip():
                yield Feature.from_chunk(chunk, wiki)

    def save(self, features):
        with self.path.open('w', encoding='utf8') as fp:
            fp.write('# -*- coding: utf-8 -*-\n')
            fp.write(self.CHUNK_SEP.join(f.as_chunk() for f in features))

    def listallwiki(self, wiki):
        fn = wiki / "List-of-all-features.md"
        currentlist = []
        for m in re.finditer(
            r"\*\s*(?P<gbid>(GB)?\d\d\d)\s+\[(?P<feature>[^]]+)]\((?P<link>[^)]+)\)",
            fn.read_text(encoding='utf-8')
        ):
            res = m.groupdict()
            if not res['gbid'].startswith('GB'):
                res['gbid'] = 'GB' + res['gbid']
            currentlist.append(res)

        active = set(f.id for f in self.iterfeatures(wiki))

        print(len(active), "active")
        print(len(currentlist) - len(active), "inactive")

        with fn.open('w', encoding='utf8') as fp:
            fp.write("Active in the GramBank feature set:\n\n")
            for spec in currentlist:
                if spec['gbid'] in active:
                    fp.write("* %(gbid)s [%(feature)s](%(link)s)\n" % spec)

            fp.write(
                "\n\nNo longer active in the GramBank feature set (for historical interest only):"
                "\n\n")
            for spec in currentlist:
                if spec['gbid'] not in active:
                    fp.write("* %(gbid)s [%(feature)s](%(link)s)\n" % spec)



"""
List active features.
"""
import textwrap

from clldutils.clilib import Table, add_format

from pygrambank.cli_util import add_wiki_repos


def register(parser):
    add_wiki_repos(parser)
    add_format(parser)


def run(args):
    contribs = {c.id: c.name for c in args.repos.contributors}
    with Table(args, '#', 'ID', 'Title', 'Patrons') as t:
        for i, f in enumerate(args.repos.features.values(), start=1):
            t.append([
                i,
                f.id,
                textwrap.shorten(f.wiki['title'], 50),
                ' '.join([contribs[abbr] for abbr in f.patrons])])



import pathlib


def add_wiki_repos(parser):
    parser.add_argument(
        '--wiki_repos',
        help="clone of grambank/grambank.wiki",
        default=pathlib.Path('../grambank.wiki'),
        type=pathlib.Path)



"""
Summary stats on the sheets in original_sheets/
"""
import collections


def run(args):
    api = args.repos

    sheets = [s for s in api.iter_sheets()]
    values = collections.Counter()
    cols = collections.Counter()

    print(api)

    print('\nSheets with BOM:')
    for sheet in sorted(sheets, key=lambda s: s.path):
        if sheet.path.read_text(encoding='utf-8').startswith('\ufeff'):
            print(sheet.path)
        for i, val in enumerate(sheet._reader(dicts=True)):
            if i == 0:
                cols.update(list(val.keys()))
            values.update([sheet.path.stem])
    print('\n{} sheets with {} rows'.format(len(values), sum(values.values())))
    print('\nMost common columns:')
    for k, v in cols.most_common(10):
        print(k, v)



"""
Download issues and associated comments from GitHub for glottobank/Grambank
"""
import os
import re

import requests
from requests.auth import HTTPBasicAuth
from clldutils import jsonlib

API = "https://api.github.com"
REPOS = "glottobank/Grambank"
LINK_PATTERN = re.compile(r'<(?P<url>[^>]+)>; rel=\"(?P<rel>[a-z]+)\"')


def register(parser):  # pragma: no cover
    parser.add_argument(
        '--user', default='xrotwang', help='GitHub user name')
    parser.add_argument(
        '--token', default=os.environ.get('GITHUB_TOKEN'), help='GitHub access token')


def run(args):  # pragma: no cover
    auth = HTTPBasicAuth(args.user, args.token)
    issues = list(iteritems(auth, '/repos/{0}/issues'.format(REPOS), state='all'))
    jsonlib.dump(issues, args.repos.issues_path, indent=4)

    res = {}
    for issue in issues:
        if issue['comments']:
            res[issue['number']] = list(iteritems(auth, issue['comments_url']))

    jsonlib.dump(res, args.repos.comments_path, indent=4)


def req(auth, url, **params):  # pragma: no cover
    if url.startswith('/'):
        url = API + url
    res = requests.get(url, params=params, auth=auth)
    links = {}
    for link in res.headers.get('Link', '').split(', '):
        match = LINK_PATTERN.match(link)
        if link:
            links[match.group('rel')] = match.group('url')
    return res.json(), links


def iteritems(auth, path, **params):  # pragma: no cover
    items, links = req(auth, path, **params)
    for item in items:
        yield item
    while links.get('next'):
        items, links = req(auth, links['next'])
        for item in items:
            yield item



"""

"""
import re
import pathlib
import collections

from nameparser import HumanName
from bs4 import BeautifulSoup as bs
from clldutils.clilib import PathType
from clldutils.markup import iter_markdown_tables

# flake8: noqa

NAME2ID = {
    'Hoju Cha': 'HC',
    'Roberto E. Herrera Miranda': 'RHE',
    'Yustinus Ghanggo Ate': 'YGA',
    'Farah El Haj Ali': 'FE',
}


def register(parser):
    parser.add_argument(
        '--html',
        type=PathType(type='file'),
        default=pathlib.Path('.') / '..' / '..' / '..' / 'glottobank.github.io' / 'people.html')


def iter_html_data(p):
    doc = bs(p.read_text(encoding='utf8'), features='html5lib')
    for div in doc.find_all('div', class_='gray-box'):
        name = div.find('strong')
        #print(name.get_text())
        #print(div.find('img')['src'] if div.find('img') else '---no photo---')
        #print(div.get_text().strip())
        yield (
            name.get_text().strip(),
            re.sub(r'\s+', ' ', div.get_text().strip()),
            div.find('img')['src'] if div.find('img') else None)


def get_row(contrib, md):
    res = collections.OrderedDict([
        ('id', contrib.id),
        ('Last name', contrib.last_name),
        ('First name', contrib.first_name),
        ('Contribution', contrib.contribution)])
    for k in ['Node', 'Status', 'Language competence', 'GitHub-username', 'email', 'photo', 'bio']:
        res[k] = md.get(k, '') or ''
    if res['photo']:
        res['photo'] = '<img src="https://glottobank.org/{}" width="100">'.format(res['photo'])
    assert not any('|' in v for v in res.values())
    return res


def run(args):
    roles = collections.Counter()
    for c in args.repos.contributors:
        roles.update([c.bio])
    for k, v in roles.most_common():
        print(k, v)
        break
    return

    bios = {r[0]: (r[1], r[2]) for r in iter_html_data(args.html)}

    # id | Last name | First name | Node | Status | Language competence | GitHub-username | email
    header, rows = next(iter_markdown_tables(args.repos.path('CONTRIBUTORS_details.md').read_text(encoding='utf8')))
    rows = [dict(zip(header, row)) for row in rows]
    rows = collections.OrderedDict([(r['id'], r) for r in rows])

    contribs = args.repos.contributors
    contribs = collections.OrderedDict([(r.id, r) for r in contribs])

    assert not set(rows) - set(contribs)
    lnames = {c.last_name: c.id for c in contribs.values()}
    fnames = {'{0.first_name} {0.last_name}'.format(c): c.id for c in contribs.values()}

    bios_by_id = {}
    for name in bios:
        if name in NAME2ID:
            bios_by_id[NAME2ID[name]] = bios[name]
            continue
        hname = HumanName(name)
        if hname.last in lnames:
            bios_by_id[lnames[hname.last]] = bios[name]
            continue
        full = '{0.first} {0.last}'.format(hname)
        if full in fnames:
            bios_by_id[fnames[full]] = bios[name]
            continue
        #print('---', name)

    for i, (cid, c) in enumerate(contribs.items()):
        bio, photo = bios_by_id.get(cid, (None, None))
        if bio:
            assert '\n' not in bio and ('|' not in bio)
        md = rows.get(cid, {})
        md['bio'] = bio
        md['photo'] = photo
        contrib = get_row(c, md)
        if i == 0:
            print(' | '.join(contrib.keys()))
            print(' | '.join([' --- ' for _ in range(len(contrib.keys()))]))
        print(' | '.join(contrib.values()))



import collections

from csvw.dsv import reader
from clldutils.apilib import API
from clldutils.misc import lazyproperty
from clldutils import jsonlib
from pyglottolog.references.bibfiles import BibFile

from pygrambank.features import GB20
from pygrambank.contributors import Contributors
from pygrambank.sheet import Sheet
from pygrambank.issues import Issue


class Grambank(API):
    def __init__(self, repos, wiki=None):
        API.__init__(self, repos)
        self.wiki = wiki or self.repos.resolve().parent / 'grambank.wiki'
        self.gb20 = GB20(self.path('gb20.txt'))

    @property
    def sheets_dir(self):
        return self.repos / 'original_sheets'

    def iter_sheets(self):
        for p in sorted(self.sheets_dir.iterdir(), key=lambda i: i.stem):
            if p.is_file() and p.name not in ['.gitattributes', '.DS_Store']:
                yield Sheet(p)

    @property
    def exclude(self):
        return {
            r['Sheet']: r['Reason'] for r in
            reader(self.path('exclude_from_cldf.tsv'), dicts=True, delimiter='\t')}

    @property
    def contributors(self):
        return Contributors.from_md(self.repos / 'CONTRIBUTORS.md')

    @lazyproperty
    def bib(self):
        return {e.key: (e.type, e.fields) for e in BibFile(self.repos / 'gb.bib').iterentries()}

    @lazyproperty
    def ordered_features(self):
        return list(self.gb20.iterfeatures(self.wiki))

    @lazyproperty
    def features(self):
        return collections.OrderedDict([(f['Grambank ID'], f) for f in self.ordered_features])

    @lazyproperty
    def issues_path(self):
        return self.path('archived_discussions', 'issues.json')

    @lazyproperty
    def comments_path(self):
        return self.path('archived_discussions', 'comments.json')

    @lazyproperty
    def issues(self):
        issues = jsonlib.load(self.issues_path)
        comments = jsonlib.load(self.comments_path)
        return [Issue(issue, comments.get(str(issue['number']), [])) for issue in issues]

    def visit_feature(self, visitor):
        features = []
        for f in self.ordered_features:
            f = visitor(f, self)
            if f:
                features.append(f)
        self.gb20.save(features)



import sys
from pathlib import Path

from clldutils.clilib import ParserError, register_subcommands, get_parser_and_subparsers
from clldutils.loglib import Logging

from pygrambank.api import Grambank


def propagate_gb20(args):  # pragma: no cover
    """
    grambank --repos=PATH/TO/Grambank --wiki_repos=PATH/TO/Grambank.wiki  propagate_gb20

    Propagate changes in gb20.txt to all derived files (in Grambank repos and wiki).
    """
    api = Grambank(args.repos)
    api.gb20.listallwiki(args.wiki_repos)


def main(args=None, catch_all=False, parsed_args=None):
    import pygrambank.commands

    parser, subparsers = get_parser_and_subparsers('grambank')
    parser.add_argument(
        '--repos', help="clone of glottobank/Grambank", default=Path('.'), type=Path)
    register_subcommands(subparsers, pygrambank.commands)

    args = parsed_args or parser.parse_args(args=args)
    if not hasattr(args, "main"):  # pragma: no cover
        parser.print_help()
        return 1

    args.repos = Grambank(args.repos, wiki=getattr(args, 'wiki_repos', None))
    with Logging(args.log, level=args.log_level):
        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except ParserError as e:  # pragma: no cover
            print(e)
            return main([args._command, '-h'])
        except Exception as e:  # pragma: no cover
            if catch_all:
                print(e)
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)



from pycldf.sources import Source
from pygrambank.cldf import BibliographyMatcher
from pygrambank.sheet import Row


# Make a mock bibliography for testing

FICTIONMAN2000 = dict(
    author='Bob Fictionman',
    year='2000',
    title='Grammar of Martian')

FICTIONMAN2001 = dict(
    author='Fictionman, Bob',
    year='2001',
    title='There are days when my last name comes first')

SPACEMAN1961 = dict(
    author='Space Man, Yuriy',
    year='1961',
    title='Reflections of a man with a space in his name')

WRITEALOT2012_FIRST = dict(
    author='Steve Writealot',
    year='2012',
    title='The first thing I wrote this week')

WRITEALOT2012_ANOTHER = dict(
    author='Steve Writealot',
    year='2012',
    title='Another thing I wrote this week')

STRAUSS_MELPA = dict(
    author='Strauß, Hermann',
    title='Grammatik der Melpa Sprache',
    pages='94',
    year='no date',
    glottolog_ref_id='23344',
    hhtype='grammar_sketch',
    howpublished='Unpublished Manuscript',
    inlg='German [deu]',
    lgcode='Melpa [med]',
    macro_area='Papua')

HINZKUNZ2023 = dict(
    author='Karl Hinz and Karsten Kunz',
    year='2023',
    title="Being cited the 'alternative' way")

BIBLIOGRAPHY = {
    'Fictionman2000': ('book', FICTIONMAN2000),
    'Fictionman2001': ('book', FICTIONMAN2001),
    'Writealot2012_First': ('book', WRITEALOT2012_FIRST),
    'Writealot2012_Another': ('book', WRITEALOT2012_ANOTHER),
    'SpaceMan1961': ('book', SPACEMAN1961),
    's:Strauss:Melpa': ('misc', STRAUSS_MELPA),
    'HinzKunz2023': ('book', HINZKUNZ2023),
}

ENGLISH = 'stan1293'
BIBKEYS_BY_GLOTTOCODE = {
    ENGLISH: {
        'Fictionman2000', 'Fictionman2001', 'HinzKunz2023',
        'SpaceMan1961', 'Writealot2012_First', 'Writealot2012_Another',
    },
    'melp1238': {'s:Strauss:Melpa'},
}


def test_ignore_bare_page_numbers():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, '1234')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_sources()
    assert not bib_matcher.has_unresolved_citations()


def test_ignore_personal_comm():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Müller (p.c.)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_sources()
    assert not bib_matcher.has_unresolved_citations()
    assert row.has_valid_source()
    assert 'source not confirmed' not in row.Source_comment


def test_ignore_empty_source():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, '')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_sources()
    assert not bib_matcher.has_unresolved_citations()


def test_report_unresolved_sources():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Santa Clause (1982)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_sources()
    assert bib_matcher.has_unresolved_citations()
    unresolved = bib_matcher.get_unresolved_citations()
    expected_errors = [(('Santa Clause', '1982', ENGLISH), 1)]
    assert unresolved == expected_errors
    assert 'source not confirmed' in row.Source_comment
    assert not row.has_valid_source()


def test_report_things_that_arent_even_citations():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, r'¯\_(ツ)_/¯')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert bib_matcher.has_unresolved_citations()
    assert not bib_matcher.has_sources()
    unresolved = bib_matcher.get_unresolved_citations()
    expected_errors = [((r'¯\_(ツ)_/¯', ENGLISH), 1)]
    assert unresolved == expected_errors


def test_report_resolved_source():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Fictionman (2000)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'Fictionman2000',
        author='Bob Fictionman',
        year='2000',
        title='Grammar of Martian')
    assert first_source == (expected_source, 1)
    assert row.Source == ['Fictionman2000']
    assert row.has_valid_source()


def test_resolve_and_unresolved_source():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Fictionman (2000); Santa Clause (1982)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)

    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'Fictionman2000',
        author='Bob Fictionman',
        year='2000',
        title='Grammar of Martian')
    assert first_source == (expected_source, 1)

    assert bib_matcher.has_unresolved_citations()
    unresolved = bib_matcher.get_unresolved_citations()
    expected_errors = [(('Santa Clause', '1982', ENGLISH), 1)]
    assert unresolved == expected_errors
    assert row.has_valid_source()
    assert 'source not confirmed' not in row.Source_comment


def test_last_name_before_first_name_in_bibliography():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Fictionman (2001)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'Fictionman2001',
        author='Fictionman, Bob',
        year='2001',
        title='There are days when my last name comes first')
    assert first_source == (expected_source, 1)
    assert row.Source == ['Fictionman2001']
    assert row.Source_comment == 'Fictionman (2001)'


def test_desambiguate_based_on_title():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Writealot_another (2012)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'Writealot2012_Another',
        author='Steve Writealot',
        year='2012',
        title='Another thing I wrote this week')
    assert first_source == (expected_source, 1)
    assert row.Source == ['Writealot2012_Another']


def test_deal_with_people_who_have_spaces_in_their_last_name():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Space Man (1961)')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'Writealot2012_Another',
        author='Space Man, Yuriy',
        year='1961',
        title='Reflections of a man with a space in his name')
    assert first_source == (expected_source, 1)
    assert row.Source == ['SpaceMan1961']


def test_alternate_citation_style():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Hinz&Kunz2023, p.113')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'book', 'HinzKunz2023',
        author='Karl Hinz and Karsten Kunz',
        year='2023',
        title="Being cited the 'alternative' way")
    assert first_source == (expected_source, 1)
    assert row.Source == ['HinzKunz2023[113]']


def test_hardcoded_fallback():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Strauß (n.d.)')
    bib_matcher.add_resolved_citation_to_row('melp1238', row)
    assert not bib_matcher.has_unresolved_citations()
    assert bib_matcher.has_sources()
    first_source = bib_matcher.get_sources()[0]
    expected_source = Source(
        'misc', 's_Strauss_Melpa',
        author='Strauß, Hermann',
        title='Grammatik der Melpa Sprache',
        pages='94',
        year='no date',
        glottolog_ref_id='23344',
        hhtype='grammar_sketch',
        howpublished='Unpublished Manuscript',
        inlg='German [deu]',
        lgcode='Melpa [med]',
        macro_area='Papua')
    assert first_source == (expected_source, 1)
    assert row.Source == ['s_Strauss_Melpa']


def test_nothing_but_von_parts():
    bib_matcher = BibliographyMatcher(BIBLIOGRAPHY, BIBKEYS_BY_GLOTTOCODE)
    row = Row(None, None, 'Van der von van 2012')
    bib_matcher.add_resolved_citation_to_row(ENGLISH, row)
    assert bib_matcher.has_unresolved_citations()
    assert not bib_matcher.has_sources()
    unresolved = bib_matcher.get_unresolved_citations()
    expected_errors = [(('Van der von van', '2012', ENGLISH), 1)]
    assert unresolved == expected_errors



"""

"""
import itertools
import functools

# flake8: noqa

DATAPOINTS = """\
GB146 | Double check with Daria if there really is no concstruction of this type. | Mishchenko, Daria (p.c. 2013) | HS_loma1260.tsv
GB146 | Double check with Elena if there really is no such construction | Perekhvalskaya, Elena (p.c 2013) | HS_mwan1250.tsv
GB146 | Double check with Daria if there really is no concstruction of this type. | Mishchenko, Daria (p.c. 2013) | HS_toma1245.tsv
GB171 | Double check page source | Harvey (2002:194-195) | JLA_gaga1251.tsv
GB333 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB334 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB335 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB333 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB334 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB335 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB304 | I think so? Double check this with Jeremy. Someone killed himself a deer./ A deer got himself killed. | Saxton (1982:144) | JLA_toho1245.tsv
GB026 | Double check figure 4. | Boxwell (1990:160) | JLA_weri1253.tsv
GB158 | Acc. to Wals no, but double check Sapir (1922), then source entire grammar and delete this comment. | Sapir (1922:215-234); Nevin (1976:237-247) | JLA_yana1271.tsv
GB159 | Acc. to Wals no, but double check Sapir (1922), then source entire grammar and delete this comment. | Sapir (1922:215-234); Nevin (1976:237-247) | JLA_yana1271.tsv
GB160 | Acc. to Wals no, but double check Sapir (1922), then source entire grammar and delete this comment. | Sapir (1922:215-234); Nevin (1976:237-247) | JLA_yana1271.tsv
GB401 | Double check this page against the wiki. Labile verb mentioned. | Gijn (2006: 145) | JLA_yura1255.tsv
GB117 | double-check use of pa, p. 253 | Lee 1975:253 | MD-GR-RSI_kosr1238.tsv
GB074 | I think that all of the adpositions are postpositions, but I found something recently that looks like a preposition. I have to double-check. For now, though, let's go with no. | Robinson 2007 | MD-GR-RSI_roto1249.tsv
GB257 | Requires further research to double-check. | Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB312 | Requires further research to double-check. | Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB330 | double check this | Heath 1999:195 | RB_koyr1240.tsv
GB021 | I think. According to the grammar there are different markers for definite and indefinite non-topics ('objects'), but the examples are confusing. | Wolfenden 1971 | MD-GR-RSI_hili1240.tsv
GB118 | I think | Haiman 1980 | MD-GR-RSI_huaa1250.tsv
GB105 | [XXX CHECK] The dative is used for animates, and recipients are a priori animate I think; GR: I would think that DAT is not Direct Object | Donohue 2004 | MD-GR-RSI_kris1246.tsv
GB075 | I think so [e.g. aape in the dictionary] | Hurd and Hurd 1966; Hurd and Hurd 1970; Hurd 1977 | MD-GR-RSI_naas1242.tsv
GB053 | I'm not sure, but I think it might. It seems that everything in the fourth noun class is animate. | Robinson 2007 | MD-GR-RSI_roto1249.tsv
GB115 | I'm not sure about reflexive vs. reciprocal. I think they share the same marker. Is that a yes for this one then? Agree | Robinson 2007 | MD-GR-RSI_roto1249.tsv
GB146 | I think | NA | MD-GR-RSI_savo1255.tsv
GB160 | adjectives I think. Several word classes show reduplication. See the long note in the separate file TelefolComm.doc' | Healey 1964; Healey 1965; Healey 1965; Healey 1965; Healey 1966; Healey 1964; Healey 1977; Healey 1962; Healey 1974 | MD-GR-RSI_tele1256.tsv
GB086 | I think so, anyway, with iemo | NA | MD-GR-RSI_touo1238.tsv
GB038 | I think they are | Sagna 2008:118 | SVM_band1340.tsv
GB081 | I think it's not a real infix | Strom 2013:248 | SVM_nden1248.tsv
GB105 | [XXX CHECK] The dative is used for animates, and recipients are a priori animate I think; GR: I would think that DAT is not Direct Object | Donohue 2004 | MD-GR-RSI_kris1246.tsv
GB133 | CHECK GRAMMAR In irrealis clauses with the verb 'lack' SOV is the only order possible, and hence pragmatically unmarked p. 750. But the example given might as well be interpreted as a possessive construction: nia ina=n lalek (3sg mother=GEN lack) 'she has no mother'. In the questionnaire for ENUS II, Van Klinken describes these constructions as object incorporation giving apparent SOV order. | van Engelenhoven and Williams-van-Klinken 2005 | MD-GR-RSI_tetu1245.tsv
GB113 | MIINA CHECKS | Miina Norvik, p.c. | RK-MN_livv1244.tsv
GB123 | check | Churchward 1953:217 | HW_tong1325.tsv
GB103 | check | Rasoloson and Rubino 2005 | MD-GR-RSI_plat1254.tsv
GB134 | check | Rasoloson and Rubino 2005 | MD-GR-RSI_plat1254.tsv
GB119 | check | Van Staden 2000 | MD-GR-RSI_tido1248.tsv
GB135 | check and get back | Agbetsoamedo 2014 | HS_sele1249.tsv
GB105 | check grammar | van Engelenhoven and Williams-van-Klinken 2005 | MD-GR-RSI_tetu1245.tsv
GB027 | check grammar. | van Engelenhoven and Williams-van-Klinken 2005 | MD-GR-RSI_tetu1245.tsv
GB069 | check marking of adjectives; this may just be a matter of analysis or frequency: argument cross-referencing is optional; if affixed the adjective could always be interpreted as heading a relative clause, but it could also be the simple absence of marking. | Van Staden 2000 | MD-GR-RSI_tido1248.tsv
GB401 | check pg 73 | Churchward 1953:73 | HW_tong1325.tsv
GB026 | Double check figure 4. | Boxwell (1990:160) | JLA_weri1253.tsv
GB171 | Double check page source | Harvey (2002:194-195) | JLA_gaga1251.tsv
GB330 | double check this | Heath 1999:195 | RB_koyr1240.tsv
GB401 | Double check this page against the wiki. Labile verb mentioned. | Gijn (2006: 145) | JLA_yura1255.tsv
GB146 | Double check with Daria if there really is no concstruction of this type. | Mishchenko, Daria (p.c. 2013) | HS_loma1260.tsv
GB146 | Double check with Daria if there really is no concstruction of this type. | Mishchenko, Daria (p.c. 2013) | HS_toma1245.tsv
GB146 | Double check with Elena if there really is no such construction | Perekhvalskaya, Elena (p.c 2013) | HS_mwan1250.tsv
GB333 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB334 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB335 | Double check with team on weds | Omar 1983:203 | JLA_rung1259.tsv
GB333 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB334 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB335 | Double check with team on weds | Omar 1983:233-234 | JLA_tamb1254.tsv
GB070 | GS: You gave N. I suggest Y, since nae-ba a-m-uw-e (I-focus him-give-I indic) ‘I give him’ is distinguished from nae-‘pa na-m-iy-e (I-focus me-give-he) ‘he gives me’ not only in the verb, but also how the focal marker is formed: class N when subject, class Q when benefactor. I see no such clear example in Scott 1978, but check out p.104 #170 Oblique cases). GR: but this still doesn't count as case marking in my view; it is only a focus marker that changes. | Scott 1978:104 | MD-GR-RSI_fore1270.tsv
GB304 | I think so? Double check this with Jeremy. Someone killed himself a deer./ A deer got himself killed. | Saxton (1982:144) | JLA_toho1245.tsv
GB150 | Jeremy will check in collection of narratives | Luo 2008 | JC_yong1276.tsv
GB159 | KM: Don't think so. (I'll check on it.) | McElhanon 1970; McElhanon 1972 | MD-GR-RSI_sele1250.tsv
GB057 | MD: check that this isn't NA | Davies 1992 | MD-GR-RSI_ramo1244.tsv
GB058 | MD: check that this isn't NA | Davies 1992 | MD-GR-RSI_ramo1244.tsv
GB082 | Nick said - checking with Hedvig re Mawng productivity | Pym and Larrimore 1979 | RSI_iwai1244.tsv
GB257 | Requires further research to double-check. | Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB312 | Requires further research to double-check. | Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB313 | SY: check for pages | Barth (2019) | SY_matu1261.tsv
GB171 | Check 160 | Guo (2010:1-196) | MY_youn1235.tsv
GB074 | Check in locative and instrumental enclitics which seem to mark oblique cases, count here for prep and post. | Campbell (2006:v-115) | JLA_urni1239.tsv
GB075 | Check in locative and instrumental enclitics which seem to mark oblique cases, count here for prep and post. | Campbell (2006:v-115) | JLA_urni1239.tsv
GB150 | Check other sources. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB327 | Check pp. 38-39. Relative clause is mentioned there but the gloss is rather useless. | Piau (1985:1-147) | ER_kuma1280.tsv
GB025 | Check the example under interrogative: savai malamalapopore nake (...) | Ray (1926:373) | DB_akei1237.tsv
GB171 | Check this. | Hayward (1984:191-192) | JLA_arbo1245.tsv
GB329 | Check with Jeremy. | Gerner (2013:65) | JLA_sich1238.tsv
GB330 | Check with Jeremy. | Gerner (2013:92) | JLA_sich1238.tsv
GB331 | Check with Jeremy. | Gerner (2013:93) | JLA_sich1238.tsv
GB327 | Check! | Reichle (1981:91-94) | RHA_bawm1236.tsv
GB037 | Checking required | Yihan Chen, Rebecca Dixon and Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB252 | Checking required | Yihan Chen, Rebecca Dixon and Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB253 | Checking required | Yihan Chen, Rebecca Dixon and Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB256 | Checking required | Yihan Chen, Rebecca Dixon and Caroline Hendy p.c. (2018) | NP_wiru1244.tsv
GB265 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB266 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB270 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB273 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB275 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB276 | Not mentioned in the Deal grammar. Check others. | Deal (2010:1-460); Crook (1999:2-498); Cash (2004:1-72) | JLA_nezp1238.tsv
GB285 | Not mentioned.. requires only a question particle? Check this. | Stephen Matthews and Virginia Yip. 1994. Cantonese: A Comprehensive Grammar | JLA_yuec1235.tsv"""
DATAPOINTS = sorted((i.split('|')[-1].strip(), i.split('|')[0].strip()) for i in DATAPOINTS.split('\n'))
DATAPOINTS = {fname: set(i[1] for i in dps) for fname, dps in itertools.groupby(DATAPOINTS, lambda i: i[0])}


def fixer(fids, row):
    if row['Feature_ID'] in fids:
        fids.remove(row['Feature_ID'])
        if row['Value'] != '?':
            row['Value'] = '?'
            print('Value changed')
        row['Comment'] = ''
    return row


def run(args):
    for sheet in args.repos.iter_sheets():
        if sheet.path.name in DATAPOINTS:
            sheet.visit(
                row_visitor=functools.partial(
                    fixer,
                    DATAPOINTS[sheet.path.name],
                ))

    for gc, dps in DATAPOINTS.items():
        assert not dps



"""
Make sure a (set of) sheets contains rows for all active features.
"""
import collections

from csvw.dsv import UnicodeWriter
from clldutils.clilib import PathType

from pygrambank.sheet import Sheet
from pygrambank.cli_util import add_wiki_repos


def register(parser):
    parser.add_argument(
        'filename',
        metavar='SHEET',
        help="Path of a specific TSV file to check or substring of a filename (e.g. a glottocode)",
        default=None,
        type=PathType(type='file', must_exist=False),
    )
    add_wiki_repos(parser)


def run(args):
    if args.filename.exists():
        sheets = [Sheet(args.filename)]
    else:
        sheets = [s for s in args.repos.iter_sheets() if args.filename.name in s.path.name]
    for sheet in sheets:
        update(args, sheet)


def update(args, sheet):
    # Get rows as in "most updated sheet":
    rows = collections.OrderedDict([
        ('Feature_ID', lambda f: f.id),
        ('Feature', lambda f: f.wiki['title']),
        ('Possible Values', lambda f: f['Possible Values']),
        ('Clarifying comments', lambda f: f.wiki['Summary'].replace('\n', ' ')),
        ('Relevant unit(s)', lambda f: f['Relevant unit(s)']),
        ('Function', lambda f: f['Function']),
        ('Form', lambda f: f['Form']),
        ('Patron', lambda f: f.wiki['Patron']),
    ])
    active = {}
    for feature in args.repos.features.values():
        active[feature.id] = {k: v(feature) for k, v in rows.items()}

    # Read the current rows:
    fids, cols, rows = set(), None, []
    for i, r in enumerate(sheet.iterrows()):
        if i == 0:
            cols = list(r.keys())
        fids.add(r['Feature_ID'])
        rows.append(r)

    # Append missing features:
    added = set()
    for fid, crow in active.items():
        if fid not in fids:
            # Assemble the row, merging in matching cols from the most updated sheet.
            added.add(fid)
            rows.append(collections.OrderedDict([(col, crow.get(col, '')) for col in cols]))

    with UnicodeWriter(sheet.path, delimiter='\t', encoding='utf8') as w:
        for i, row in enumerate(rows):
            if i == 0:
                w.writerow(list(row.keys()))
            w.writerow(list(row.values()))
    if added:
        args.log.info('Sheet {}: added {} features'.format(sheet.path.name, len(added)))



"""
Do we want to recode the wiki pages in CLDF markdown?
It would seem that this may add a nice IGT example/reference collection.

Also, a global references section for all wiki articles might be nice?
"""
import collections

from fuzzywuzzy import fuzz


def iter_examples(s):
    #
    # FIXME: must also parse the last mentioned Glottocode before each example!
    #
    def split_examples(ss):
        for ee in ss.split('\n\n'):
            ee = ee.strip()
            if ee:
                yield ee

    in_example, ex = False, []
    for line in s.split('\n'):
        line = line.strip()

        if line.startswith('```'):
            if in_example:
                yield from split_examples('\n'.join(ex))
                ex = []
            in_example = not in_example
        else:
            if in_example:
                ex.append(line)


def run(args):
    sections = collections.Counter()
    refs = []
    for feature in args.repos.features.values():
        sections.update(list(feature.wiki))
        for line in feature.wiki.get('References', '').split('\n'):
            line = line.strip()
            if line:
                refs.append(line)
        for line in feature.wiki.get('Further reading', '').split('\n'):
            line = line.strip()
            if line:
                refs.append(line)

        continue
        for ex in iter_examples(feature.wiki['Examples']):
            if len(ex.split('\n')) != 3:
                print(feature.id)
                break

    c, prev = 0, ''
    for ref in sorted(refs):
        if fuzz.token_set_ratio(prev, ref) < 100:
            print(ref)
            print('')
            c += 1
        prev = ref



"""
Write merged sheets for conflict resolution
"""
import pathlib
import itertools

from clldutils.clilib import PathType
from csvw.dsv import UnicodeWriter

from pygrambank.sheet import Row


def register(parser):
    parser.add_argument(
        '--outdir', default=pathlib.Path('.'), type=PathType(type='dir', must_exist=False))


def grouped_rows(sheets):
    for fid, vals in itertools.groupby(
        sorted(
            itertools.chain(*[[(r, s) for r in s.iterrows()] for s in sheets]),
            key=lambda i: (i[0]['Feature_ID'], i[1].path.name)),
        lambda r: r[0]['Feature_ID'],
    ):
        yield fid, [(r, s) for r, s in vals if r['Value']]


class Warnings:
    def __init__(self):
        # Accumulator for warning messages:
        self.messages = ''

    def __call__(self, msg, lineno, level=None, row_=None, **kw):
        # This will be called within Sheet.valid_row when problems are detected.
        self.messages += '\n{}:{}: {}'.format(level, row_['Feature_ID'], msg)


def run(args):
    api = args.repos
    if not args.outdir.exists():
        args.outdir.mkdir()

    # Iterate over groups of sheets for the same glottocode:
    for gc, sheets in itertools.groupby(
            sorted(api.iter_sheets(), key=lambda s: s.glottocode),
            lambda s: s.glottocode):
        sheets = list(sheets)
        if len(sheets) == 1:
            continue

        with UnicodeWriter(args.outdir / '{}.tsv'.format(gc), delimiter='\t') as w:
            w.writerow([
                'Feature_ID',
                'Value',
                'Conflict',
                'Classification of conflict',
                'Select',
                'Sheet',
                'Source',
                'Contributed_Datapoint',
                'Comment',
                'Warnings',
            ])
            # Iterate over rows grouped by feature ID:
            for fid, rows in grouped_rows(sheets):
                conflict = len(set(r[0]['Value'] for r in rows if r[0]['Value'])) > 1
                for row, sheet in rows:
                    # See what "grambank check" would do with the row:
                    warnings = Warnings()
                    sheet.valid_row(row, api, log=warnings)
                    w.writerow([
                        fid,
                        row['Value'],
                        str(conflict),
                        '',
                        '',
                        sheet.path.stem,
                        row.get('Source', ''),
                        ' '.join(Row.from_dict(row).contributed_datapoint),
                        row.get('Comment', ''),
                        warnings.messages,
                    ])




def test_api_contributors(api):
    assert api.contributors[0].photo == 'abc'
    assert len(api.contributors[0].roles) == 2


def test_features(api, wiki):
    assert len(api.features) == 9
    assert len(api.features['GB021'].domain) == 2
    f = api.features['GB021']
    assert f.description
    assert f.name
    assert f.patrons
    assert f.id
    assert 'Documentation' in api.features['GB020'].description


def test_bib(api):
    assert len(api.bib) == 6


def test_sheets(api):
    assert len(list(api.iter_sheets())) == 2



import re
import itertools
import collections
from termcolor import colored

import attr
from csvw import dsv

from pygrambank.bib import iter_authoryearpages, mismatch_is_fatal


def check_feature_dependencies(rows):
    values = {
        r.Feature_ID: r
        for r in rows
        if r.Feature_ID and r.Value}

    def _value(feat):
        row = values.get(feat)
        return row.Value if row else None

    def _comment(feat):   # pragma: nocover
        row = values.get(feat)
        return row.Comment if row else None

    def _require_comment(feat, reason):
        if not _comment(feat):
            errors.append('{} must have a comment if {}'.format(feat, reason))

    errors = []

    if (_value('GB408')
        == _value('GB409')
        == _value('GB410')
        == '0'
    ):
        errors.append("GB408, GB409, and GB410 can't all be 0")  # pragma: nocover

    if (_value('GB131')
        == _value('GB132')
        == _value('GB133')
        == '0'
    ):
        errors.append("GB131, GB132, and GB133 can't all be 0")  # pragma: nocover

    if (_value('GB083')
        == _value('GB084')
        == _value('GB121')
        == _value('GB521')
        == '0'
        and _value('GB309') == '1'
    ):
        errors.append(
            "GB309 can't be 1 if GB083, GB084, GB121 and GB521 are all 0")  # pragma: nocover

    if (_value('GB333')
        == _value('GB334')
        == _value('GB335')
        == _value('GB336')
        == '0'
    ):
        reason = 'GB333, GB334, GB335, and GB336 are all 0'
        for feat in ('GB333', 'GB334', 'GB335', 'GB336'):  # pragma: nocover
            _require_comment(feat, reason)

    for feat in (
        'GB026', 'GB303', 'GB320', 'GB166', 'GB197', 'GB129', 'GB285', 'GB336',
        'GB260', 'GB165', 'GB319'
    ):
        if _value(feat) == '1':
            _require_comment(feat, reason='it is coded 1')  # pragma: nocover

    if (_value('GB265')
        == _value('GB266')
        == _value('GB273')
        == 0
    ):
        reason = 'GB265, GB266, and GB273 are all 0'
        _require_comment('GB265', reason)
        _require_comment('GB266', reason)
        _require_comment('GB273', reason)

    if (_value('GB072')
        == _value('GB073')
        == _value('GB074')
        == _value('GB075')
        == 0
    ):
        reason = 'GB072, GB073, GB074, and GB075 are all 0'
        _require_comment('GB074', reason)
        _require_comment('GB075', reason)

    if _value('GB155') == 1 and _value('GB113') == 0:
        reason = 'GB155 is 1 and GB113 is 0'
        _require_comment('GB155', reason)
        _require_comment('GB113', reason)

    return errors


@attr.s
class Source:
    author = attr.ib()
    year = attr.ib()
    pages = attr.ib()
    in_title = attr.ib()

    @property
    def key(self):
        return self.author, self.year, self.in_title


@attr.s
class Row:
    Feature_ID = attr.ib()
    Value = attr.ib()
    Source = attr.ib()
    Comment = attr.ib(default=None)
    contributed_datapoint = attr.ib(
        default=attr.Factory(list),
        converter=lambda s: re.findall('[A-Z]+(?=[^A-Z]|$)', s) if s else [])
    Source_comment = attr.ib(default=None)

    @classmethod
    def from_dict(cls, d):
        fields = list(attr.fields_dict(cls).keys())
        kw = {}
        for k, v in d.items():
            if ('ontributed' in k) and ('atapoint' in k):
                k = 'contributed_datapoint'
            if k in fields:
                kw[k] = v
        return cls(**kw)

    @property
    def sources(self):
        return [Source(*ayp) for ayp in iter_authoryearpages(self.Source)]

    def has_valid_source(self):
        """Return True if row has at least one matched source."""
        sources = self.Source
        source_string = self.Source_comment
        return (
            isinstance(sources, list)
            and (bool(sources) or not mismatch_is_fatal(source_string)))


class Sheet(object):
    """
    Processing workflow:
    """
    name_pattern = re.compile(
        r'(?P<coders>[A-Z]+(-[A-Z]+)*)_(?P<glottocode>[a-z0-9]{4}[0-9]{4})\.tsv$')

    def __init__(self, path):
        match = self.name_pattern.match(path.name)
        assert match, 'Invalid sheet name: {0}'.format(path.name)
        self.path = path
        self.coders = match.group('coders').replace('CB-PE-AS', 'HunterGatherer').split('-')
        self.glottocode = match.group('glottocode')
        self._rows = None

    def __str__(self):
        return str(self.path)

    def metadata(self, glottolog):
        languoid = glottolog.languoids_by_glottocode[self.glottocode]
        if languoid.level.name == 'dialect':
            for _, lgc, level in reversed(languoid.lineage):
                if level.name == 'language':
                    break
        else:
            lgc = languoid.id
        language = glottolog.languoids_by_glottocode[lgc]
        if not language.macroareas:
            print('--- no macroareas: {}'.format(self.glottocode))
        return dict(
            level=languoid.level.name,
            lineage=[lang[1] for lang in languoid.lineage],
            Language_level_ID=language.id,
            # Macroareas are assigned to language level nodes:
            Macroarea=language.macroareas[0].name if language.macroareas else '',
            Latitude=languoid.latitude if languoid.latitude else language.latitude,
            Longitude=languoid.longitude if languoid.longitude else language.longitude,
            Family_name=languoid.lineage[0][0] if languoid.lineage else None,
            Family_level_ID=languoid.lineage[0][1] if languoid.lineage else None,
        )

    def _reader(self, **kw):
        return dsv.reader(self.path, delimiter='\t', encoding='utf-8-sig', **kw)

    def iterrows(self):
        if self._rows is None:
            self._rows = []
            for row in self._reader(dicts=True):
                self._rows.append(row)
                yield row
        else:
            for row in self._rows:
                yield row

    def visit(self, row_visitor=None):
        """
        Apply `row_visitor` to all rows in a sheet.

        :param row_visitor:
        :return: Pair of `int`s specifying the number of rows read and written.
        """
        if row_visitor is None:
            row_visitor = lambda r: r  # noqa: E731
        rows = list(self.iterrows())
        count = 0
        with dsv.UnicodeWriter(self.path, delimiter='\t', encoding='utf8') as w:
            for i, row in enumerate(rows):
                if i == 0:
                    w.writerow(list(row.keys()))
                res = row_visitor(row)
                if res:
                    w.writerow(list(row.values()))
                    count += 1
        # Make sure calling iterrows again will re-read from disk:
        self._rows = None
        return (len(rows), count)

    def valid_row(self, row, api, lineno=None, log=None, features=None):
        fid = row.get('Feature_ID')
        if not fid:
            return False
        res = True
        if not re.match('GB[0-9]{3}|(GBDRS.+)|TE[0-9]+|TS[0-9]+$', fid):
            if row.get('Value'):
                if log:
                    log('invalid Feature_ID: {0}'.format(fid),
                        lineno=lineno,
                        level='ERROR',
                        row_=row)
            res = False
        if fid not in api.features:
            return False
        if row.get('Value'):
            if row['Value'] != '?' and row['Value'] not in api.features[row['Feature_ID']].domain:
                if log:
                    log('invalid value: {0}'.format(row['Value']), lineno=lineno, row_=row)
                res = False
        else:
            res = False

        if row['Value'] and not row['Source']:
            if log:
                log('value without source', lineno=lineno, level='WARNING', row_=row)
            res = False
        if row['Source'] and not row['Value']:
            if log:
                log('source given, but no value', lineno=lineno, level='WARNING', row_=row)
            res = False
        if row['Comment']:
            if (log is not None
                and 'check' in row['Comment'].lower()
                and 'HG' not in row['Comment']
                and 'checked by coder' not in row['Comment'].lower()
                and 'check by coder' not in row['Comment'].lower()
                and 'wrong import' not in row['Comment'].lower()
                and 'checked by GB coder' not in row['Comment'].lower()
            ):
                log('comment contains string "check"', lineno=lineno, level='WARNING', row_=row)
            if not row['Value']:
                if log:
                    log('comment given, but no value', lineno=lineno, level='WARNING', row_=row)
                res = False
        if row['Feature_ID'] in (features or set()):
            if log:
                log('duplicate value for feature {0}'.format(
                    row['Feature_ID']), lineno=lineno, level='ERROR', row_=row)
            res = False
        return res

    def check(self, api, report=None):
        def log(msg, row_=None, level='ERROR', lineno=-1):
            msg = [
                self.path.stem,
                level,
                "{}".format(lineno if lineno != -1 else '?'),
                row_['Feature_ID'] if row_ else '',
                msg]
            print(colored('\t'.join(msg), color='red'))
            if report is not None:
                report.append(msg)

        # Check the header:
        empty_index = []
        for line, row in enumerate(self._reader()):
            if line == 0:
                for col in ['Feature_ID', 'Value', 'Comment', 'Source']:
                    if col not in row:
                        log('missing column {0}'.format(col), lineno=line)
                for j, c in enumerate(row):
                    if not c:
                        empty_index.append(j)
                if len(set(row)) != len(row):
                    dupes = collections.Counter([h for h in row if row.count(h) > 1])
                    log('duplicate header column(s) %r' % dupes, lineno=line)
            else:
                if not empty_index:
                    break
                for j in empty_index:
                    if row[j]:
                        log('non-empty cell with empty header: {0}'.format(row[j]),
                            lineno=line,
                            level='WARNING')

        res, nvalid, features, comments = [], 0, set(), 0
        for line, row in enumerate(self.iterrows(), 2):  # +2 to get line number correct
            if self.valid_row(row, api, lineno=line, log=log, features=features):
                nvalid += 1
                comments += 1 if row['Comment'] else 0
            features.add(row['Feature_ID'])
            res.append(row)

        if comments < 25:
            log('Less than 25 datapoints have comments', level='WARNING')

        try:
            for gbid, rows in itertools.groupby(
                sorted(res, key=lambda r: r['Feature_ID']), lambda r: r['Feature_ID']
            ):
                rows = list(rows)
                if len(rows) > 1:
                    # A feature is coded multiple times! If the codings are inconsistent, we raise
                    # an error, otherwise the first value takes precedence.
                    if len(set(r['Value'] for r in rows)) > 1:
                        log('inconsistent multiple codings: {0}'.format([r['Value'] for r in rows]))
        except:  # pragma: no cover  # noqa
            for row in res:
                if not row['Feature_ID']:
                    print(row)
                    break
            print(self.path)
            raise

        for msg in check_feature_dependencies(self.iter_row_objects(api)):
            log(msg)  # pragma: nocover

        return nvalid

    def itervalues(self, api):
        try:
            for row in self.iterrows():
                if self.valid_row(row, api):
                    yield row
        except:  # pragma: no cover # noqa: E722
            print(self.path)
            raise

    def iter_row_objects(self, api):
        for row in self.itervalues(api):
            yield Row.from_dict(row)



import re

AUTHOR_PATTERNS = [
    re.compile(p) for p in [
        r"(?P<lastname>[^,]+),\s((?P<jr>[JS]r\.|[I]+),\s)?(?P<firstname>[^,]+)$",
        r"(?P<firstname>[^{][\S]+(\s[A-Z][\S]+)*)\s"
        r"(?P<lastname>([a-z]+\s)*[A-Z\\\\][\S]+)(?P<jr>,\s[JS]r\.|[I]+)?$",
        r"(?P<firstname>\\{[\S]+\\}[\S]+(\s[A-Z][\S]+)*)\s"
        r"(?P<lastname>([a-z]+\s)*[A-Z\\\\][\S]+)(?P<jr>,\s[JS]r\.|[I]+)?$",
        r"(?P<firstname>[\s\S]+?)\s\{(?P<lastname>[\s\S]+)\}(?P<jr>,\s[JS]r\.|[I]+)?$",
        r"\{(?P<firstname>[\s\S]+)\}\s(?P<lastname>[\s\S]+?)(?P<jr>,\s[JS]r\.|[I]+)?$",
        r"(?P<lastname>[A-Z][\S]+)$",
        r"\{(?P<lastname>[\s\S]+)\}$",
        r"(?P<lastname>[aA]nonymous)$",
        r"(?P<lastname>\?)$",
        r"(?P<lastname>[\s\S]+)$",
    ]
]

REGEX_ONLY_PAGES = re.compile(r"[\d+;\s\-etseqpassim.]+$")

PARTIAL_REGEX_PAGES = r"(?:\:\s*(?P<p>[\d\,\s\-]+))?"
PARTIAL_REGEX_YEAR = r"(?:\d\d\d\d|no date|n.d.|[Nn][Dd])"

REGEX_FULL_SOURCE = re.compile(
    r"^(?P<a>[^,]+),[^(\d]+[\s(](?P<y>"
    + PARTIAL_REGEX_YEAR + r")\s*"
    + PARTIAL_REGEX_PAGES + r"\)?")

PARTIAL_REGEX_CAPITALS = r'ÅA-Z\x8e\x8f\x99\x9aÉ'
REGEX_SOURCE = re.compile(
    r"(?P<a>(?<![^\s(])["
    + PARTIAL_REGEX_CAPITALS + r"vd][a-z]*\D*[^\d,.])\.?\s\(?(?P<y>"
    + PARTIAL_REGEX_YEAR + r")"
    + PARTIAL_REGEX_PAGES + r"\)?")

# Gwynn&Krishnamurti1985, p.144
REGEX_ALTERNATIVE_STYLE_SOURCE = re.compile(
    r"^(?P<a>[A-Z][a-zA-Z&]+)(?P<y>[0-9]{4}),\s+p\.\s*(?P<p>[\d,\s\-]+(?:ff?\.)?)$")

REGEX_NAME_PARTS = re.compile(r"\s+|(d\')(?=[A-Z])")
REGEX_CAPITALISED_NAME_PART = re.compile(
    r'\[?[{}]'.format(PARTIAL_REGEX_CAPITALS))

REGEX_BRACKETED_CODES = re.compile(r"\[([a-z0-9]{4}[0-9]{4}|[a-z]{3}|NOCODE_[A-Z][^\s\]]+)\]")
REGEX_LGCODE_SEPARATOR = re.compile(r"[,/]\s?")
REGEX_ISO_CODE = re.compile(r"[a-z][a-z][a-z]$|NOCODE_[A-Z][^\s\]]+$")

HHTYPE_RANKING_OLD = [
    'grammar',
    'grammar sketch',
    'dictionary',
    '(typological) study of a specific feature',
    'phonology',
    'text',
    'new testament',
    'wordlist',
    'comparative-historical study',
    'some very small amount of data/information on a language',
    'endangered language',
    'sociolinguistically oriented',
    'dialectologically oriented',
    'handbook/overview',
    'ethnographic work',
    'bibliographically oriented',
    'unknown']

HHTYPE_NEW_TYPES = {
    'grammar': 'grammar',
    'grammar sketch': 'grammar_sketch',
    'dictionary': 'dictionary',
    '(typological) study of a specific feature': 'specific_feature',
    'phonology': 'phonology',
    'text': 'text',
    'new testament': 'new_testament',
    'wordlist': 'wordlist',
    'comparative-historical study': 'comparative',
    'some very small amount of data/information on a language': 'minimal',
    'endangered language': 'endangered language',
    'sociolinguistically oriented': 'socling',
    'dialectologically oriented': 'dialectology',
    'handbook/overview': 'overview',
    'ethnographic work': 'ethnographic',
    'bibliographically oriented': 'bibliographical',
    'unknown': 'unknown',
}

HHTYPE_RANKING = [HHTYPE_NEW_TYPES[hhtype] for hhtype in HHTYPE_RANKING_OLD]
HHTYPE_PRIORITIES = {
    hhtype: len(HHTYPE_RANKING) - i
    for i, hhtype in enumerate(HHTYPE_RANKING)}


def get_hhtypes(bibentry):
    """Return publication types of a bibliography entry.

    Based on the classification in the `hh` bibliography in glottolog.
    """
    hhtype_field = bibentry.get('hhtype', 'unknown')
    hhtype_field = re.sub(r' \([^)]*\)', '', hhtype_field)
    return re.split(r'[;,]\s?', hhtype_field)


def hhtype_priority(hhtype):
    """Return priority of a publication type."""
    return HHTYPE_PRIORITIES.get(hhtype, 0)


def split_off_lowercase_nameparts(first_name):
    """Split off lower-case name prefixes (de, von, van, etc.)."""
    parts = [s for s in REGEX_NAME_PARTS.split(first_name) if s]
    for i, name_part in enumerate(parts):
        if REGEX_CAPITALISED_NAME_PART.match(name_part):
            return (parts[:i], parts[i:])
    else:  # pragma: nocover
        return (), ()


def move_von_part_to_last_name(author):
    """Sometimes the 'von' parts of a name end up in the 'firstname' field.

    This fixes that.
    """
    if 'firstname' not in author:
        return author
    r = {}
    (lower, upper) = split_off_lowercase_nameparts(author['firstname'])
    r['lastname'] = (' '.join(lower).strip() + ' ' + author['lastname']).strip()
    r['firstname'] = ' '.join(upper)
    if author.get('jr'):
        r['jr'] = author['jr']
    return r


def parse_single_author(author_string):
    """Parse a single author from a BibTeX string.

    Return a dictionary {'firstname': XXX, 'lastname': YYY, 'jr': ZZZ}.
    """
    for p in AUTHOR_PATTERNS:
        match = p.match(author_string)
        if match:
            return move_von_part_to_last_name(match.groupdict())
    return None


def parse_authors(authors_string):
    """Parse the 'authors' field of a BibTeX entry."""
    authors_string = authors_string.replace(' & ', ' and ')
    authors = authors_string.split(' and ')
    authors = [parse_single_author(a.strip()) for a in authors]
    authors = list(filter(None, authors))
    return authors


def bibkey_authors(bibkey):
    """Generator yielding author names as encountered in a citation key."""
    # A citation key as used in hh.bib!
    if ':' in bibkey:
        bibkey = bibkey.split(':')[1]

    current_name = ''
    for char in bibkey:
        # For keys of the form "Meier2018" we stop at the start of year.
        if char.isdigit():
            break
        elif char.isupper() and current_name not in ('Mac', 'De'):
            if current_name:
                yield current_name
            current_name = char
        else:
            current_name += char
    if current_name:
        yield current_name


def lgcodestr(lgcstr):
    """Parse language codes in the `lgcode` field of a BibTeX entry.

    example:
        lgcode = {Scots [scot1243], English [eng]}
    """
    lgs = REGEX_BRACKETED_CODES.findall(lgcstr)
    if lgs:
        return lgs

    parts = [p.strip() for p in REGEX_LGCODE_SEPARATOR.split(lgcstr)]
    codes = [p for p in parts if REGEX_ISO_CODE.match(p)]
    return codes


def prioritised_bibkeys(bibkeys, bibliography_entries):
    """Return the bibkeys with the highest priority.

    e.g. prefer full grammars over grammar sketches.
    """
    bibkey_rankings = {}
    for bibkey in bibkeys:
        bibentry = bibliography_entries[bibkey][1]
        hhtypes = get_hhtypes(bibentry)
        hhtype_ranking = max(map(hhtype_priority, hhtypes))
        written_in_english = lgcodestr(bibentry.get('inlg', "")) == ['eng']
        bibkey_rankings[bibkey] = hhtype_ranking, written_in_english
    highest_rank, _ = max(
        (ranking, bibkey)
        for bibkey, ranking in bibkey_rankings.items())
    prioritised_bibkeys = {
        bibkey
        for bibkey, ranking in bibkey_rankings.items()
        if ranking == highest_rank}
    return prioritised_bibkeys


def mismatch_is_fatal(source_string):
    """Return True iff. an unmatched source string constitutes an error.

    i.e., ignore stuff like Smith (personal communication).
    """
    # Note: In theory this code could be combined into one big boolean
    # expression but I doubt that will it any more readable...
    if REGEX_ONLY_PAGES.match(source_string):
        # TODO: Maybe find a way to warn about this
        # print(
        #     'PAGEONLY:',
        #     '[%s] default source:%s' % (glottocode, source_string),
        #     glottocode)
        return False
    elif (
        'p.c' in source_string
        or 'personal communication' in source_string
        or 'pers comm' in source_string
        or 'pers. comm' in source_string
        or 'ieldnotes' in source_string
        or 'ield notes' in source_string
        or 'forth' in source_string
        or 'Forth' in source_string
        or 'ubmitted' in source_string
        or 'o appear' in source_string
        or 'in press' in source_string
        or 'in prep' in source_string
        or 'in prog' in source_string
        or source_string.startswith('http')
    ):
        return False
    else:
        return True


def iter_authoryearpages(source_string):
    """Parse the `Source` field of a data sheet.

    Returns an iterator of tuples with the following components:
     * Author name
     * Year of publication
     * Page number
     * Word from title
    """
    for citation_string in source_string.replace("), ", "); ").split(";"):
        if "p.c." in citation_string:
            continue
        condensed = False
        citation_string = citation_string.strip()

        m = REGEX_FULL_SOURCE.search(citation_string)
        if not m:
            m = REGEX_SOURCE.search(citation_string)

        if not m:
            condensed = True
            m = REGEX_ALTERNATIVE_STYLE_SOURCE.search(citation_string)

        if m:
            author, year, pages = m.groups()
            if condensed:
                author = author.replace('&', ' and ')
            wft = author.find("_")
            if wft != - 1:
                word_from_title = author[wft + 1:].lower()
                author = author[:wft]
            else:
                word_from_title = ''
            yield (author, year, pages.strip() if pages else pages, word_from_title)



import re
import warnings
import itertools
import collections

from csvw import dsv
import openpyxl
import xlrd

GB_COLS = collections.OrderedDict([
    ("Language_ID", ["iso-639-3", "Language", "Glottocode", "glottocode"]),
    ("Feature_ID", ["GramBank ID", "Grambank ID", r"\* Feature number", "Grambank"]),
    ("Value", []),
    ("Source", []),
    ("Comment", ["Freetext comment"]),
    ("Feature Domain", ["Possible Values"]),
])


def normalized_feature_id(s):
    if s.isdigit():
        s = "GB" + str(s).zfill(3)
    elif s.startswith("GB") and len(s) != 5:
        s = "GB" + str(s[2:]).zfill(3)
    return s


def normalize_comment(s):
    """
    Normalize comments, turning things like "????" into "?".

    :param s: The original comment
    :return: The normalized comment as string
    """
    if s:
        if set(s) == {'#'}:
            return
        if set(s) == {'?'}:
            return '?'
        return s


def normalized_value(v):
    if v in {
        '?',
        '??',
        'n/a',
        'N/A',
        'n.a.',
        'n.a',
        'N.A.',
        'N.A',
        '-',
        'NODATA',
        '? - Not known'
        '*',
        "*",
        '\\',
        'x',
    }:
        return '?'  # pragma: no cover
    return v


def _normalized_row(row):
    for k in row:
        row[k] = row[k].strip() if row[k] else row[k]

    # Normalize column names:
    if ('Grambank ID' in row) and ('Feature_ID' in row):
        row['Feature'] = row.pop('Feature_ID')  # pragma: no cover

    for col, aliases in GB_COLS.items():
        if col not in row:
            for k in list(row.keys()):
                if k in aliases:
                    row[col] = row.pop(k)
                    break
            else:
                row[col] = ''

    # Normalize colum values:
    row['Language_ID'] = None
    row['Feature_ID'] = normalized_feature_id(row['Feature_ID'])
    row['Value'] = normalized_value(row.get('Value'))
    row['Comment'] = normalize_comment(row['Comment'])
    return row


def _read_excel_value(x):
    if x is None:
        return ""
    if isinstance(x, float):
        return '{0}'.format(int(x))
    return '{0}'.format(x).strip()


def iter_tsv(fname):
    for row in dsv.reader(fname, delimiter='\t', encoding='utf-8-sig', dicts=True):
        yield _normalized_row(row)


def iter_xlsx(fname):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        sheet = openpyxl.load_workbook(str(fname), data_only=True).active

        header = None
        empty_rows = 0
        empty_cols, data = [], []
        skip_cols = set()
        for i, row in enumerate(sheet.rows):
            if i == 0:
                for j, c in enumerate(row):
                    # There's a couple of sheets with > 10,000 columns, labeled with "Column<no>".
                    # We cut these out to reduce TSV bloat.
                    if re.match('Column[0-9]+$', '{0}'.format(c.value)):  # pragma: no cover
                        skip_cols.add(j)
            row = [_read_excel_value(c.value) for j, c in enumerate(row) if j not in skip_cols]
            if set(row) == {''}:  # pragma: no cover
                empty_rows += 1
                if empty_rows > 1000:
                    # There's a couple of sheets with > 100,000 (mostly empty) rows.
                    # After encountering more than 1,000, we stop reading.
                    break
                else:
                    continue
            if header is None:
                header = row
                empty_cols = [i for i, c in enumerate(header) if not c]
            else:
                assert len(header) == len(row), 'Header and row length mismatch'
                data.append(row)
                yield _normalized_row(collections.OrderedDict(zip(header, row)))

        for i in empty_cols:  # pragma: no cover
            assert all(not bool(r[i]) for r in data), \
                'Empty header for non-empty column: {0}'.format(i)


def iter_xls(fname):
    wb = xlrd.open_workbook(str(fname))
    rows_by_sheetname = collections.defaultdict(list)
    for sheet in wb.sheets():  # We read all sheets in the workbook.
        for row in range(sheet.nrows):
            rows_by_sheetname[sheet.name].append([
                _read_excel_value(sheet.cell_value(row, col))
                for col in range(sheet.ncols)])
    # Now select the proper sheet:
    rows = None
    if len(rows_by_sheetname) > 1:
        for sheetname in ["GramBank", '"Empty" GramBank Sheet']:
            if sheetname in rows_by_sheetname:
                rows = rows_by_sheetname[sheetname]
                break
    else:  # pragma: no cover
        rows = list(rows_by_sheetname.values())[0]
    assert all(bool(c) for c in rows[0]), 'Empty column header'
    for row in rows[1:]:
        yield _normalized_row(collections.OrderedDict(zip(rows[0], row)))


def iter_csv(fname):
    def read(encoding):
        with fname.open(encoding=encoding) as csvfile:
            line = csvfile.readline()
            delimiter = ','
            if ';' in line and ((',' not in line) or (line.index(';') < line.index(','))):
                delimiter = ';'
        for row in dsv.reader(
                fname,
                delimiter=delimiter,
                quotechar='"',
                doublequote=True,
                encoding=encoding,
                dicts=True):
            yield _normalized_row(row)

    try:
        res = list(read('utf-8-sig'))
    except UnicodeDecodeError:  # pragma: no cover
        res = list(read('cp1252'))

    for r in res:
        yield r


def write_tsv(in_, out_, glottocode):
    rows = list({
        '.xlsx': iter_xlsx,
        '.xls': iter_xls,
        '.csv': iter_csv,
        '.tsv': iter_tsv,
    }[in_.suffix](in_))

    i = 0
    with dsv.UnicodeWriter(out_, delimiter='\t') as w:
        for i, row in enumerate(rows):
            if i == 0:
                w.writerow(list(row.keys()))
            row['Language_ID'] = glottocode
            w.writerow(list(row.values()))
    return i


def iterunique(insheets, verbose=False):
    """
    For languages which have been coded multiple times, we pick out the best sheet.
    """
    # Sort sheets by glottocode and number of values, then group:
    for gc, sheets in itertools.groupby(
            sorted(insheets, key=lambda s: (s[0].glottocode, -len(s[1]), s[0].path.stem)),
            lambda s: s[0].glottocode):
        sheets = list(sheets)
        if len(sheets) == 1:
            yield sheets[0]
        else:
            if verbose:
                print('\nSelecting best sheet for {0}'.format(gc))
            for i, (sheet, values) in enumerate(sheets):
                if verbose:
                    print('{0} dps: {1} sheet {2}'.format(
                        len(values), 'choosing' if i == 0 else 'skipping', sheet.path.stem))
                if i == 0:
                    yield (sheet, values)



"""
Lookup the references given in the `Source` column of a sheet in Glottolog hh.bib or the Grambank bib.
"""
import pathlib
import collections

from termcolor import colored
from cldfcatalog import Catalog

from pygrambank.sheet import Sheet
from pygrambank.cldf import BibliographyMatcher, GlottologGB
from pygrambank.bib import lgcodestr


def register(parser):
    parser.add_argument(
        'glottolog',
        metavar='GLOTTOLOG',
        help="clone of glottolog/glottolog",
        type=pathlib.Path,
    )
    parser.add_argument(
        'sheets',
        type=pathlib.Path,
        nargs='+',
    )
    parser.add_argument(
        '--glottolog-version',
        default=None,
        help="tag to checkout glottolog/glottolog to",
    )


def run(args):
    if args.glottolog_version:  # pragma: no cover
        with Catalog(args.glottolog, args.glottolog_version) as glottolog:
            run_(args, glottolog.dir)
    else:  # pragma: no cover
        run_(args, args.glottolog)


def run_(args, glottolog):  # pragma: no cover
    grambank = args.repos
    sheets = [Sheet(sh) for sh in args.sheets]

    # FIXME: code duplication with cldfbench
    print('Reading language data from Glottolog...')
    glottolog = GlottologGB(glottolog)
    languoids_by_ids = glottolog.languoids_by_ids
    descendants = glottolog.descendants_map

    print('Loading bibliography...')
    bibliography_entries = {}
    bibliography_entries.update(glottolog.bib('hh'))
    bibliography_entries.update(grambank.bib)

    bibkeys_by_glottocode = collections.defaultdict(set)
    for key, (typ, fields) in bibliography_entries.items():
        for lang_id in lgcodestr(fields.get('lgcode') or ''):
            if lang_id in languoids_by_ids:
                glottocode = languoids_by_ids[lang_id].id
                if glottocode in descendants:
                    for cl in descendants[glottocode]:
                        bibkeys_by_glottocode[cl].add(key)
                else:
                    print('---non-language', lang_id)

    for sheet in sheets:
        glottocode = sheet.glottocode

        print(colored(
            '\nSource look-up for sheet {}...\n'.format(sheet.path),
            attrs=['bold']))

        bib_matcher = BibliographyMatcher(
            bibliography_entries, bibkeys_by_glottocode)
        for sheet_row in sheet.iter_row_objects(args.repos):
            bib_matcher.add_resolved_citation_to_row(
                sheet.glottocode, sheet_row)

        print(colored('Resolved sources:', attrs=['bold']))
        for source, occurrence_count in bib_matcher.get_sources():
            print('{}\t{}\t{}'.format(occurrence_count, source.id, source))

        if bib_matcher.has_unresolved_citations():
            print()
            print(colored('Unresolved sources:', attrs=['bold']))
            for spec, occurrences in bib_matcher.get_unresolved_citations():
                if len(spec) == 3:
                    author, year, _ = spec
                    print('{}\t{} ({})'.format(occurrences, author, year))
                elif len(spec) == 2:
                    source_string, _ = spec
                    print('{}\t{}'.format(occurrences, source_string))
                else:  # pragma: nocover
                    # theoretically unreachable
                    print('{}\t{}'.format(occurrences, spec))
            if bibkeys_by_glottocode.get(glottocode):
                print()
                print(colored('Available sources:', attrs=['bold']))
                for bibkey in bibkeys_by_glottocode[glottocode]:
                    type_, fields = bibliography_entries[bibkey]
                    author = fields.get('author') or fields.get('editor') or '-'
                    year = fields.get('year') or '-'
                    print('{}\t{}\t{}'.format(
                        colored(bibkey, color='blue'),
                        type_,
                        colored('{} {}'.format(author, year), attrs=['bold'])))
            print()
            print(colored('FAIL', color='red'))
        else:
            print()
            print(colored('OK', color='green'))



"""
Describe a (set of) sheets.

This includes checking for correctness - i.e. the functionality of `grambank check`.
While references will be parsed, the corresponding sources will **not** be looked up
in Glottolog (since this is slow). Thus, for a final check of a sheet, you must run
`grambank sourcelookup`.
"""
import collections

from termcolor import colored
from clldutils.clilib import PathType

from pygrambank.sheet import Sheet


def register(parser):
    parser.add_argument(
        'filename',
        metavar='SHEET',
        help="Path of a specific TSV file to check or substring of a filename (e.g. a glottocode)",
        default=None,
        type=PathType(type='file', must_exist=False),
    )
    parser.add_argument(
        '--columns',
        default=False,
        action='store_true',
        help='List columns of the sheet',
    )
    parser.add_argument(
        '--read-only',
        default=False,
        action='store_true',
        help='Do not roundtrip',
    )


def run(args):
    if args.filename.exists():
        sheets = [Sheet(args.filename)]
    else:
        sheets = [s for s in args.repos.iter_sheets() if args.filename.name in s.path.name]
    for sheet in sheets:
        describe(args, sheet)
        if not args.read_only:
            sheet.visit(lambda r: r)


def describe(args, sheet):
    rows = list(sheet.iter_row_objects(args.repos))

    def head(s):
        print()
        print(colored(s + ':', attrs=['bold']))

    head('Path')
    print(str(sheet.path))

    head('Check')
    sheet.check(args.repos)

    head('Dimensions')
    print('{} rows'.format(len(list(sheet.iterrows()))))
    for row in sheet.iterrows():
        print('{} columns'.format(len(row)))
        break

    head('Values')
    stats = collections.Counter(r.Value for r in rows)
    for k, v in stats.most_common():
        print('{}\t{}'.format(k, str(v).rjust(3)))
    print(colored('\t{}'.format(str(sum(stats.values())).rjust(3)), attrs=['bold']))

    refs = collections.Counter()
    for row in rows:
        for ref in row.sources:
            refs.update([ref.key])

    head('Sources')
    for (author, year, in_title), v in refs.most_common():
        print('{} {}{}:\t{}'.format(
            author,
            year,
            ' ({})'.format(in_title) if in_title else '',
            str(v).rjust(3)))

    head('Coders')
    coders = collections.Counter()
    for row in rows:
        coders.update(set(sheet.coders).union(row.contributed_datapoint))
    coders_by_id = {c.id: c.name for c in args.repos.contributors}

    for k, v in coders.most_common():
        if k in coders_by_id:
            print('{}\t{}:\t{}'.format(k, coders_by_id[k], str(v).rjust(3)))
        else:
            args.log.error(colored(
                'Unknown coder: "{}"'.format(k),
                color='red'))  # pragma: no cover

    if args.columns:
        head('Columns')
        for row in sheet.iterrows():
            for col in row:
                print(col)
            break



"""

"""
import functools
import itertools

from clldutils.misc import slug

# flake8: noqa

DATAPOINTS = """
GB068 | Edelsten & Lijongwa (2010: 40, 67-71) | Compare adjectival concord and verbal concord. | ndam1239
GB069 | Edelsten & Lijongwa (2010: 40, 67-71) | Compare adjectival concord and verbal concord. | ndam1239
GB069 | Hansson (2017:893) | Adjectives are for the most part stative verbs, but when used attributively they must be preceded either by adjectival prefix 'jɔ-', or by a reduplicated syllable of the head noun; the two constructions denote temporary and inherent properties respectively. | akha1245
GB069 | Hargus (2007:319-321, 557-563) | Property concepts are expressed in a variety of different ways; some are considered descriptive or neuter verbs and inflect as such, and some fall into one of 3 subclasses of adjective. Only the latter can be used attributively | babi1235
GB069 | Cook (2013:395) | Only two property words - k'an 'new' and jed 'old' - can be used attributively; all other descriptive property concepts are expressed as stative verbs. | chil1280
GB069 | Solnit (1997:5-6, 186-190, 253-258) | There is no true class of 'adjectives' in Kayah, rather stative or property verbs. In order to be used attributively, they (whether property word or verbal modifier) must be placed within a postposed relative clause. Moreover, there is no verbal inflection in the language, and thus no morphology which the two might share. | east2342
GB069 | Goossen (1995:19, 106); Williams (2009:115); Young & Morgan (1987:189) | There is no independent class of adjectives in Navajo, only stative or 'adjectival' verbs; they cannot be used attributively. | nava1243
GB069 | Woodbury (2018:143) | There are no adjectives, only stative verbs expressing property meaning; these cannot be used attributively. | onon1246
GB069 | Cook (1984:67, 100, 157) | There is no seperate class of adjectives. Property concepts are expressed either by 'neuter' verbs, or by nominal qualifiers, derivational affixes which are often derived from the corresponding neuter verb themselves. | sars1236
GB068 | Kamanda-Kola 2003: 279-289 | Stative verbs can be derived from adjectives by changing tone. | mono1270
GB069 | McWhorter and Good (2012:107) | Since there is almost no verbal morphology, I code 0 because of the differential reading of verbs and property words under reduplication. | sara1340
GB068 | de Vries (1993:31, 36) | Verbs can, however, be productively derived from adjectives (p. 14) | komb1274
GB068 | van de Kerke (2009:316-317) | There are some tendencies but, e.g., verbal negation is not allowed | leco1242
GB069 | Sibomana (2008:28-29) | Though many are derived from verbs | zarm1239
GB069 | Robinson 2008: 135 | Adjectives that do not inflect for aspect exist | dupa1235
GB069 | Johnston 1980: 161 | Very small set of core adjectives does not behave like verbs | naka1262
GB069 | Shultz (1997:18-19) | There are very few (if any) true adjectives. They do not have any verbal features. Other property words which are based on verb roots behave also differently than verbs when used attributively. | komc1235
GB069 | Blood (1999:15-16); Mbibeh (1996:89) | The few existing adjectives are derived from stative verbs. These derived forms do not behave like verbs. | okuu1243
GB069 | Valentine (2001:342-347, 592) | All property concepts can act like verbs and then they require the same morphology as intransitive verbs. However, many concepts can also be expressed as so-called prenouns, deriving a new word by prefixing to the noun. These are considered as core adjectives here. | otta1242
GB068 | Smith (1979:84-85) | Both take no morphology. | seda1262
GB069 | Smith (1979:84-85) | Both take no morphology. | seda1262
GB069 | Todd (1970:64-68, 112, 144-152, 201, 203, 239-240) | Some concepts are expressed as verbs, but not all of them. | seve1240
GB069 | Andrews (1994:87-89, 93-95); Boling (1981:118) | The concepts which can be expressed as so-called prenouns are considered as core adjectives here. All other property concepts require verbal morphology. | shaw1249
GB069 | Engelkemier (2010:18-21, 25) | There is very limited verbal morphology (reciprocal, causative). Adjectives are considered stative verbs and seem to have no marking at all in attributive position. | west2397
GB069 | Segerer, G (2013 p.c.) | It could be analyzed as a participle | kera1248
GB068 | Fofana 2002:121 | There are however a few adjectival roots (""be black and white"" for example). | pula1263
GB069 | Doris Richter. (forthcoming) A Grammar of Mbembe.:134-151 | They are often derived from verbs in a participle form. | tigo1236
GB068 | Nashipu (2005:125-128) | True 'non-derived' adjectives do not appear to be used predicatively. | baba1264
GB068 | Sommer (2010:291); Zorc (1967:71) | There is no disction between verbal and adjectival roots. When there is the prefix 'ma-' and the translation is predicative. Adnominal property words take the linker 'nga'. | bant1288
GB068 | Singh (2000:144-146, 152-153, 201) | ""Adjective is derived from the verbal roots. In this case, prefix e- is used as a formative particle and the nominalizer -pa is suffixed to the verbal root"" p.145 | chot1239
GB069 | Singh (2000:144-146, 152-153, 201) | ""Adjective is derived from the verbal roots. In this case, prefix e- is used as a formative particle and the nominalizer -pa is suffixed to the verbal root"" p.145 | chot1239
GB068 | Toba (1984:27, 30-31) | There is no separate class of adjectives - they are derived from verbs. | khal1275
GB069 | Toba (1984:27, 30-31) | There is no separate class of adjectives - they are derived from verbs. | khal1275
GB069 | Foulkes (1915:218-219) | There is no bound marking (but there is a preceding particle 'kō' which creates a ""substantive"".) | ngas1240
GB068 | Crowther (1864:12-13, 15) | No adjectival or verbal inflection. | nupe1254
GB069 | Crowther (1864:11-13, 15); Kawu (2002:209-214) | No adjectival or verbal inflection. | nupe1254
GB068 | Tiwari (1971:76, 88, 123, 127-130) | No definable class of adjectives, only verbs convey property concepts. | rong1266
GB069 | Tiwari (1971:76, 88, 123, 127-130) | No definable class of adjectives, only verbs convey property concepts. | rong1266
GB068 | Kobayashi 2012:29 | no TAM marking; person marking is possible (p.29) | saur1249
GB068 | Barkman (2014:60, 81-83, 115-118) | No separate word class for adjectives is found, but attributive verbs function as such. | para1302
GB069 | Barkman (2014:60, 81-83, 115-118) | No separate word class for adjectives is found, but attributive verbs function as such. | para1302
GB068 | Shelden (1989:xxi-149) | Galela use a kind of relative clause headed by a stative verb, or also uses participial forms of stative verbs. There are no adjectives in this language. | gale1259
GB069 | Shelden (1989:xxi-149) | Galela use a kind of relative clause headed by a stative verb, or also uses participial forms of stative verbs. There are no adjectives in this language. | gale1259
GB069 | Cai (2016:110-111, 117) | Neither take marking. | honi1244
GB068 | Omar 1983:256 | Take an adjectival prefix to mark the predicative function of the adjective. | lotu1250
GB069 | Jenny (2005:109) | Verbal infix turns stative verbs into attributive adjectives. | monn1252
GB068 | Codrington & Palmer 1896:xvii | Not always | mota1237
GB069 | Codrington & Palmer 1896:xvii | Not always | mota1237
GB069 | Robson (2011:110) | Most adj are derived from verbs, very small number are not. They do not seem to take markings like verbs do however. | njan1240
GB068 | Beltrame (1877: 15-16) | Adjectives take no inflection. | asoa1238
GB069 | Beltrame (1877: 15-16) | Adjectives take no inflection. | asoa1238
GB068 | De Rendinger (1949: 181) | There is no adjectival morphology. | bolg1251
GB069 | De Rendinger (1949: 181) | There is no adjectival morphology. | bolg1251
GB068 | Lesage 2020:264 | They are always derived from verbs. They do not take the morphology nouns take in predicative position, so in that sense they're more akin to verbs than to nouns. But they are still radically different from verbs. They are not marked for tense, aspect and negation, for instance. | kamm1249
GB069 | Lesage 2020: 173-177 | They are always derived from verbs. So the stem is always a verb. They do not take any verbal inflection, however. | kamm1249
GB069 | Lord 2016: 44, 34 | Although there is no discussion on verbs used attributively, see example 57. | yora1241
GB068 | Bon 2014:1-628 | Acc. to Bon (2014), property words are statif verbs, however, there is no verbal morphology on either verbs or adjectives. | bulo1242
GB069 | Bon 2014:1-628 | Acc. to Bon (2014), property words are statif verbs, however, there is no verbal morphology on either verbs or adjectives. | bulo1242
GB068 | Cammenga 2002:499-500 | Only one example of a color term in a copula construction, in this case without verbal morphology. All other terms without examples and glosses. | gusi1247
GB068 | Jenny and Mccormick 2015:539 | Although property concepts are expressed by verbs, they do not take any morphology. | oldm1242
GB069 | Jenny and Mccormick 2015:539 | Although property concepts are expressed by verbs, they do not take any morphology. | oldm1242
GB068 | Svantesson 1988:101 | Adjectives are stative verbs but there is no morphology. | uuuu1243
GB069 | Svantesson 1988:101 | Adjectives are stative verbs but there is no morphology. | uuuu1243
GB069 | Prentice 1971: 32-33, 190-193 | The relator used with adjectives can be omitted. | timu1262
GB068 | Bennett, C. E. (1908:34-43) | some core adjectives do e.s. Rubere (cf. Rau 2009: esp. 114-115, 123-125) | lati1261
GB069 | Dillon 1994:23 | Attributive adjectives have no affixes. | tata1257
GB069 | Paudyal (2015:72-75,162) | ""extremely few descriptive adjectives in Chintang"" (Paudyal 2015:73): | chhi1245
GB069 | Field 1997:267-272 | “[adjectives] do not take verbal morphology unless they are derived as verbs” (p271) | dong1285
GB069 | Kratochvíl 2007:92 | In constrast to other verbs, stative verbs do not require embedding in a relative clause with the intersective linker ba (CSEQ). | abui1241
GB068 | Holzknecht 1986:106-108 | mentions two Types of adj: Type 1 includes shape, age, dimension and value = 'bare'adjectives, attributively and predicatively; Type II have some value, but further propensities, and these are stative verbs in predicative position. | adze1240
GB069 | Holzknecht 1986 | This answer holds for Type I; Type II require verb-like treatment = nominalization by Participle /-an/. | adze1240
GB068 | Conrad and Wogiga 1991:33, 37 | Class 6 = Stative verbs have object suffix for S; Non-statives: S prefix | buki1249
GB068 | Green 1987; Glasgow 1981; Glasgow 1980; Green 2003 | Adjectives agree with their modifier in class, case, person, number and gender. Agreement markers used are the same as for verbs. Adjectives differ from verbs in that they do not take TAM marking and adjectives take case markers (Green:8). | bura1267
GB068 | Bunye & Yap 1971a; Bunye & Yap (1971b); Payne 1994; Valkama 2000 | Differ from verbs in not taking voice marking | cebu1242
GB069 | Bunye & Yap 1971a; Bunye & Yap (1971b); Payne 1994; Valkama 2000 | Differ from verbs in not taking voice marking | cebu1242
GB068 | Scott 1978 | but all utterances require final MOOD, e.g. IND /=e/. GS: You gave N. Since –e is a mood suffix that completes an utterance on verbs, and on non-verbs when they end (usually short) utterances, my answer would be Y. (You already referenced Scott 1978:84).\|\|GR: Yes, but no Tense & P/N! | fore1270
GB069 | Dougherty 1983:92 | apparently not; no TAM marking (p. 92) | futu1245
GB068 | Wolfenden 1971 | Distinct from verbs in that they don't take the voice marking. | hili1240
GB069 | Wolfenden 1971 | Distinct from verbs in that they don't take the voice marking. | hili1240
GB069 | Haiman 1980:266 | most adjectives are analyzed as nouns | huaa1250
GB068 | Donohue and San Roque 2004:39, 41 | adjectives, unlike verbs, begin with consonant, and lack S prefix when predicative | kris1246
GB068 | Franklin 1971; Franklin 1978; Franklin 1981; Yarapea 2001 | áá rúdu 'man is short'= 'It is a short man' or 'He is a man who is short'; Such predicative use may use an existential verb as pia 'sit' (p.87; See also Franklin 1981). KF: There are adjectives that act like verbs in predicative positions:. Aa epe-ae = The man who is good... | west2599
GB068 | Vuillermet 2012: 544 | there are predicative adjectives which usually take special morphology. The author mentions a dozen occurences whre predicative adjectives display unexpected verbal morphology, but it is not a general characteristic of property words. | esee1248
GB069 | Dai (2009:71-72, 76-77) | What is labelled as adjectives are mostly stative verbs which form relative clauses with an auxiliary particle in order to be used attributively; same with regular verbs. Though there are also examples in which these property words act like true adjectives and do not require the formation of a relative clause. | enuu1235
GB069 | Dai and Li (2007:127, 138) | Attributive verbs must form a relative clause. This appears optional with property words. Relative clauses are formed with the genitive particle ta55 and therefore not considered as morphological. | lash1243
GB069 | Chen, Wan and Lai (1986:41, 43-44); Deepadung, Rattanapitak and Buakaw (2015:1076, 1092) | Most property words appear to be stative verbs though not explicitly described as such in Chen et al (1986). There is a group of adjectives that form a relative clause when used attributively (like verbs) but there are also examples of adjectives that simply occur after the noun. The formation of the relative clauses however, is syntactical (word order) and does involve any morphological markers. Deepadung et al (2015) however describes the property words as adjectival stative verbs. | ruch1235
GB068 | Kossmann (2013:129) | There are no adjectives. | ghad1239
GB069 | Kossmann (2013:129) | There are no adjectives. | ghad1239
GB069 | el Hannouche (2008:66, 73) | Adjectives inflect for feminine gender and plural; participles have no inflection for these categories. | ghom1257
GB069 | Jauncey (2011:184, 275) | There is no morphologically distinguishable adjective word class in Tamambo adjectival meaning is expressed by verbs. | malo1243
GB068 | Ryding (2005:240, 438) | Adjectives in predicative use agree with the subject in gender and number. Verbal predicates also agree in person and inflect for TAME. | stan1318
GB068 | Schnell (2011:67, 72) | There are only three adjectives in Vera'a and those are not considered a grammatically definable class. Properties are expressed by property verbs. | vera1241
GB069 | Schnell (2011:67,72) | There are only three adjectives in Vera'a and those are not considered a grammatically definable class. Properties are expressed by property verbs. | vera1241
GB068 | Juldasev (1981:169); Poppe (1964:87-88) | Nominals are either unmarked used predicatively or take a predicative suffix indicating person and number as well as a plural number agreement suffix. Poppe (1964:87): there is no number agreement. Number of the S/A argument is not marked on the verb -> number marked on the adjective cannot be considered “acting like verbs” | bash1264
GB068 | Sumbatova (2012:39, 41, 56-57) | Examples do not provide evidence. | land1256
GB069 | Sumbatova (2012:26-61); Sumbatova (1999:533) | Examples do not provide evidence. | land1256
GB068 | Blackings (2003:106) | “There is no morphological diflerence between attributive and predicative adjectives.” | madi1260
GB069 | Blackings (2003:106) | “There is no morphological diflerence between attributive and predicative adjectives.” | madi1260
GB069 | Payne and Payne (2013:283) | There is no distinct class of adjectives in Panare; some descriptive are nouns, some are adverbs (see Payne and Payne 2013:281-282) | enap1235
GB068 | Reichle (1981:85) | Adjectives are described as ""adjective-verbs"" and must be followed by modifier tak in a predicative construction. Transitive verbs in predicate position do not require this. | bawm1236
GB068 | Rubino (2005:311, 313, 316) | There are also stative verbs but predicative adjectives receive no stative marking. | bayb1234
GB069 | Blust (1999:321-365) | Adjectives are stative verbs in this language and so aren't used attributively. | kulo1237
GB068 | Ding (1998:92) | Verbs in predicative position are labelled 'descriptive verbs' but do not have a different form. | nort2723
GB068 | Fabre (2016) | There is no adjectival word class. | niva1238
GB069 | Fabre (2016) | There is no adjectival word class. | niva1238
GB069 | Palancar (2009) | There is no adjectival word class. | quer1236
GB068 | Moser (2004: 845ff) | there is no attributive copula, most adnominal attribution is expressed through relative clauses and predicative attribution through stative verbs. The small group of core adjectives do not seem to take verbal morphology when used predicatively. | seri1257
GB068 | Einarsson 1976:116 | However, they can take on verbal morphology (""X greens""). | icel1247
GB068 | Hantgan 2014:37f | Not entirely clear from the examples. There are also verbal adjectives, however they differ from adjectives in their formal behaviour. | tief1242
GB068 | Dempwolff n.d.:37-39 | Predicative adjectives can occasionally take TAM suffixes, however, this creates a change in meaning consistent with the TAM marking; in most cases, predicative adjectives are treated like predicative nominals. | geda1237
GB069 | Meng 2018:85, 119-122 | Verbs do not take affixes/'morphological treatment' | taul1251
GB068 | Naden 1988:29-30 | Adjectives cannot be used predicatively | fare1241
GB069 | Betty Snell 1978; Snell, Betty E. 1998; José Pío Aza 2005 | only some, there are different adj classes.Datapoint inherited from Danielsen’s work in the database of South American Indigenous Language Structures (SAILS). | mach1267
GB069 | Constenla Umaña 1991:190 | only in negative clauses | wapi1253
GB068 | Hellwig 2011:9 | no adjectives | goem1240
GB069 | Hellwig 2011:9 | no adjectives | goem1240
GB068 | Innes 1967:105 | no adjectives | mend1266
GB069 | Innes 1967:105 | no adjectives | mend1266
GB068 | Hellenthal 2010:210 | adjectives are derived from verbs | shek1245
GB069 | Hellenthal 2010:210 | adjectives are derived from verbs | shek1245
GB068 | Dum-Tragut (2009: 116); Zorc (1995: 265-268) | Adjectives can be used in attributive and predicative functions. In both functions they do not agree with the noun(s) in number and case. | nucl1235
GB069 | Dum-Tragut (2009: 116) | Adjectives can be used in attributive and predicative functions. In both functions they do not agree with the noun(s) in number and case. | nucl1235
GB068 | Haspelmath 1993:116 | The (verbal) predicative suffix -da used to be attached to adjectives, but this is no longer the case in the modern standard language. | lezg1247
GB069 | Peterson (2011:126-127) | Although there is no separate lexical class of adjectives, contentive morphemes require a different morphosyncactic treatment as verbs in attributive position. | khar1287
GB069 | Nagaya 2012: 175-200 | Similar to verbs, adjectival verbs need to be nominalized when used attributively. However, these nominalizers are not used productively by regular verbs anymore. Regular verbs use a nominalizer borrowed from Indonesian most of the time. | lewo1244\
""".split('\n')


def split_dp(line):
    parts = line.split('|')
    # Comments may contain "|", so we have to be a bit careful when parsing markdown table rows:
    return (
        parts[0].strip(),  # What comes before the first "|"
        parts[1].strip(),  # Between first and second "|"
        '|'.join(parts[2:-1]).strip(),  # Between second and last "|"
        parts[-1].strip()  # After last "|"
    )


DATAPOINTS = [split_dp(row) for row in DATAPOINTS if row]
DATAPOINTS = sorted(DATAPOINTS, key=lambda i: (i[-1], i[0]))
# Turn datapoints into nested dicts. This allows us to keep track of everything we found by deleting
# the corresponding entry.
DATAPOINTS = {
    # glottocode: {feature_id: comment ...}
    gc: {dp[0]: dp[2] for dp in dps}
    for gc, dps in itertools.groupby(DATAPOINTS, lambda i: i[-1])}


def fixer(datapoints, row):
    """
    :param datapoints: `dict` mapping feature IDs of datapoints to fix for a glottocode to comments.
    :param row: A `dict` of (column, value) pairs, representing a row in a sheet.
    """
    if row['Feature_ID'] in datapoints:
        # Here's a feature that might need fixing. Check if the comment is as expected:
        # (Since there may be more sheets per glottocode, we must make sure to not fix in the
        # wrong sheets.)
        # Sluggifying the comment strings makes for a simple "fuzzy" matching.
        if slug(row['Comment']) == slug(datapoints[row['Feature_ID']]):
            assert row['Value'] == '0'
            row['Value'] = '?'
            row['Comment'] = ''
            # We delete the datapoint to be able to check for completeness of the fixing lateron.
            del datapoints[row['Feature_ID']]
    return row


def run(args):
    for sheet in args.repos.iter_sheets():
        if sheet.glottocode in DATAPOINTS:
            # A sheet that might contain datapoints to fix!
            sheet.visit(
                row_visitor=functools.partial(
                    fixer,
                    DATAPOINTS[sheet.glottocode],  # We pass in the dict of features to fix.
                ))

    # Make sure we fixed all datapoints:
    for gc, dps in DATAPOINTS.items():
        assert not dps



"""

"""
import pathlib

from csvw.dsv import reader


def register(parser):  # pragma: no cover
    parser.add_argument('groupings')


def run(args):  # pragma: no cover
    for p in pathlib.Path(args.groupings).glob('*.csv'):
        groupings = {r['Feature_ID']: r for r in reader(p, dicts=True)}
        new_features = []
        for feature in args.repos.ordered_features:
            for k, v in groupings[feature.id].items():
                if k != 'Feature_ID':
                    if k in feature:
                        pass
                    else:
                        feature[k] = v
            new_features.append(feature)
    args.repos.gb20.save(new_features)



import collections
from itertools import groupby
import re
import unicodedata

import pyglottolog
from clldutils.misc import lazyproperty
from pycldf.sources import Source

from pygrambank import bib


MANUAL_SOURCE_MATCHES = {
    ('Strauß', 'n.d.', 'melp1238'): 's:Strauss:Melpa',
    ('Thurston', '1987', 'kuan1248'): 'hvw:Thurston:NBritain',
    ("Z'Graggen", '1969', 'geda1237'): 'hvld:Zgraggen:Madang',
    ('LeCoeur and LeCoeur', '1956', 'daza1242'): 's:LeCoeurLeCoeur:Teda-Daza',
    ('Lindsey', '2018', 'agob1244'): 'Lindsey2018',
    ('Lee', '2005a', 'mada1285'): 'Lee2005a',
    ('Lee', '2005b', 'mada1285'): 'Lee2005b',
}

VON_PREFIXES = [
    'De', 'Da', 'Van', 'Von', 'Van den', 'Van der', 'Von der', 'El', 'De la',
    'De', "d'"]


def undiacritic(s):
    return ''.join(
        ' ' if c.isspace() else c
        for c in unicodedata.normalize('NFD', s)
        if c.isascii() and (c.isalnum() or c.isspace()))


def clean_bibkey(key):
    """Remove colons in bibliography key."""
    return key.replace(':', '_').replace("'", "")


class BibliographyMatcher:
    """Object for resolving citations in a datasheet.

    How to use it:

     1. Loop over the rows of a `Sheet` object.
     2. Call `add_resolved_citation_to_row` on each row.  This part is very
        side-effect-heavy:
        1. The method replaces the Source string in the row with the key of
           the bibliography entry and adds the original citation to the
           `Source_comment` field.  The row is mutated in-place.
        2. The object keeps an record of any matched sources.
        3. The object also keeps a record for all citations that failed to
           match.
     3. Get matched sources via the `get_sources()` method.
     4. For error handling, get all failed matches via the
        `get_unresolved_citations()` method.
    """

    def __init__(self, bibliography_entries, bibkeys_by_glottocode):
        self._unresolved_citations = collections.Counter()
        self._sources = collections.OrderedDict()
        self._source_occurrences = collections.Counter()
        self._bibliography_entries = bibliography_entries
        self._bibkeys_by_glottocode = bibkeys_by_glottocode

        # Note: Caching the family names of the authors because parsing them
        # turned out to be quite expensive.
        self._parsed_author_cache = {}

    def has_sources(self):
        """Return True iff. there are citations which could be resolved."""
        return bool(self._sources)

    def get_sources(self):
        """Return a list of tuples (source, occurrence_count)."""
        return [
            (source, self._source_occurrences[bibkey])
            for bibkey, source in self._sources.items()]

    def has_unresolved_citations(self):
        """Return True iff. there are citations which couldn't be resolved."""
        return bool(self._unresolved_citations)

    def get_unresolved_citations(self):
        """Return a list of tuples (citation, occurrence_count)."""
        return self._unresolved_citations.most_common()

    def _parse_authors(self, bibkey):
        """Return set of family names of authors for a bibkey (cached)."""
        if bibkey in self._parsed_author_cache:
            return self._parsed_author_cache[bibkey]
        else:
            bibentry = self._bibliography_entries[bibkey][1]
            bibauthors = bibentry.get('author') or bibentry.get('editor') or ''
            family_names = {
                undiacritic(x['lastname'])
                for x in bib.parse_authors(bibauthors)}
            family_names.update(bib.bibkey_authors(bibkey))
            self._parsed_author_cache[bibkey] = family_names
            return family_names

    def _bibkeys_from_citation(
        self, author, year, pages, word_from_title, glottocode
    ):
        """Return bibliography keys corresponding to a citation."""

        citation_lastname = undiacritic(bib.parse_authors(author)[0]['lastname'])
        for name_part in re.split(r'[\s,.\-]+', citation_lastname):
            if name_part.strip() and name_part[0].isupper() and name_part not in VON_PREFIXES:
                citation_firsttoken = name_part
                break
        else:
            citation_firsttoken = None
        word_from_title_norm = word_from_title.replace('_', ' ')

        bibkeys = []
        for bibkey in self._bibkeys_by_glottocode[glottocode]:
            bibentry = self._bibliography_entries[bibkey][1]
            bibtitle = bibentry.get('title', '').lower()
            family_names = self._parse_authors(bibkey)

            if year not in bibentry.get('year', ''):
                continue
            if word_from_title_norm and word_from_title_norm not in bibtitle:
                continue
            if not any(
                citation_firsttoken in re.split(r'[\s,.\-]+', lastname)
                for lastname in family_names
            ):
                continue

            bibkeys.append(bibkey)

        return bibkeys

    def add_resolved_citation_to_row(self, glottocode, sheet_row):
        """Destructively add citations to the row of a datasheet.

        The `BibliographyMatcher` keeps track of all matched or unmatched
        citations it encounters.  Those can be retrieved using the
        `get_sources` and `get_unresolved_citations` methods.
        """
        if not sheet_row.Source:
            return

        matched_refs = set()
        unmatched_refs = set()

        source_string = sheet_row.Source
        authoryears = list(bib.iter_authoryearpages(source_string))
        if not authoryears and bib.mismatch_is_fatal(source_string):
            unmatched_refs.add((source_string, glottocode))

        for author, year, pages, word_from_title in authoryears:
            bibkeys = self._bibkeys_from_citation(
                author, year, pages, word_from_title, glottocode)
            if bibkeys:
                # FIXME: only yield at most one match!?
                matched_refs.update(
                    (bibkey, pages)
                    for bibkey in bib.prioritised_bibkeys(
                        bibkeys, self._bibliography_entries))
            elif (author, year, glottocode) in MANUAL_SOURCE_MATCHES:
                matched_refs.add(
                    (MANUAL_SOURCE_MATCHES[(author, year, glottocode)], pages))
            else:
                unmatched_refs.add((author, year, glottocode))

        # output: record unsucessful matches

        self._unresolved_citations.update(unmatched_refs)

        # output: record successful matches

        matched_refs = sorted(matched_refs, key=lambda r: (r[0], r[1] or ''))
        matched_refs = [
            (key, clean_bibkey(key), pages) for key, pages in matched_refs]

        for old_bibkey, new_bibkey, _ in matched_refs:
            if new_bibkey not in self._sources:
                type_, fields = self._bibliography_entries[old_bibkey]
                self._sources[new_bibkey] = Source(type_, new_bibkey, **fields)
            self._source_occurrences[new_bibkey] += 1

        # output: update sheet row

        pages_for_bibkey = collections.OrderedDict()
        for new_bibkey, key_refs in groupby(matched_refs, lambda ref: ref[1]):
            if new_bibkey not in pages_for_bibkey:
                pages_for_bibkey[new_bibkey] = set()
            pages_for_bibkey[new_bibkey].update(
                pages for _, _, pages in key_refs if pages)

        sheet_row.Source = [
            '{}{}'.format(
                new_bibkey,
                '[{}]'.format(','.join(sorted(pages))) if pages else '')
            for new_bibkey, pages in pages_for_bibkey.items()]
        if not sheet_row.Source and bib.mismatch_is_fatal(source_string):
            sheet_row.Source_comment = '{} (source not confirmed)'.format(
                source_string)
        else:
            sheet_row.Source_comment = source_string


class GlottologGB(object):
    """
    A custom facade to the Glottolog API.
    """
    def __init__(self, repos):
        self.api = repos if isinstance(repos, pyglottolog.Glottolog) \
            else pyglottolog.Glottolog(repos)

    def bib(self, key):
        """
        Retrieve entries of a Glottolog BibTeX file.

        :param key: filename stem of the BibTeX file, e.g. "hh" for "hh.bib"
        :return: dict mapping citation keys to (type, fields) pairs.
        """
        return {
            e.key: (e.type, e.fields)
            for e in self.api.bibfiles['{0}.bib'.format(key)].iterentries()}

    @lazyproperty
    def languoids(self):
        return list(self.api.languoids())

    @lazyproperty
    def languoids_by_glottocode(self):
        return {lang.id: lang for lang in self.languoids}

    @lazyproperty
    def descendants_map(self):
        res = collections.defaultdict(list)
        for lang in self.languoids:
            res[lang.id].append(lang.id)
            if lang.lineage:
                for _, gc, _ in lang.lineage:
                    res[gc].append(lang.id)
        return res

    @lazyproperty
    def languoids_by_ids(self):
        """
        We provide a simple lookup for the three types of identifiers for a Glottolog languoid,
        where hid takes precedence over ISO 639-3 code.
        """
        res = {}
        for lang in self.languoids:
            res[lang.id] = lang
            if lang.iso:
                res[lang.iso] = lang
        for lang in self.languoids:
            if lang.hid:
                res[lang.hid] = lang
        return res



"""
Clean mojibake from comments in sheets
"""
from ftfy import fix_encoding
from termcolor import colored


def visit(row):  # pragma: no cover
    if row['Comment']:
        fixed = fix_encoding(row['Comment'])
        if fixed != row['Comment']:
            print('{}\n{}'.format(
                colored('---' + row['Comment'], 'red'),
                colored('+++' + fixed, 'green'),
            ))
            row['Comment'] = fixed
    return True


def run(args):  # pragma: no cover
    for sheet in args.repos.iter_sheets():
        sheet.visit(row_visitor=visit)



"""
Fix sheet(s) by applying a python lambda function to all its rows.
"""
import pathlib

from pygrambank.sheet import Sheet


def register(parser):
    parser.add_argument(
        'lambda_',
        metavar='lambda',
        help='Python lambda accepting row-dict as sole argument, '
             'e.g. "lambda r: r.update(Source=r[''Source''].replace(''), '', ''); '')) or r"')
    parser.add_argument(
        'sheets',
        metavar='SHEETS',
        nargs='+',
    )


def run(args):
    for sheet in args.sheets:
        p = pathlib.Path(sheet)
        if p.exists() and p.is_file():
            sheet = p.name
        sheet = Sheet(args.repos.sheets_dir / sheet)
        args.log.info('fixing {0}'.format(sheet.path))
        read, written = sheet.visit(row_visitor=eval(args.lambda_))
        args.log.info('{0} rows read, {1} rows written'.format(read, written))



"""

"""
import shutil
import collections


def run(args):
    gcs = collections.Counter(
        p.stem.split('_')[-1]
        for p in args.repos.path('original_sheets').glob('*.tsv'))
    for p in args.repos.path('conflicts').glob('*.tsv'):
        if p.stem in gcs:
            if gcs[p.stem] == 1:
                print(p)
                shutil.move(p, args.repos.path('conflicts_resolved', p.name))



import pytest

from pygrambank import util


@pytest.mark.parametrize(
    'fname,nrows',
    [
        ('Other Coder_Lang [NOCODE_lang].xlsx', 2),
        ('Some One_Some Lang [NOCODE_xyz].csv', 0),
        ('The Other Coder_Language [iso].tsv', 1),
        ('Yet Another Coder_New Lang [abcd1234].xls', 0),
    ]
)
def test_write_tsv(api, tmp_path, fname, nrows):
    assert nrows == util.write_tsv(
        api.path('obsolete_sheets', fname), tmp_path / 't.tsv', 'abcd1234')



import shutil
import pathlib
import argparse
import collections

import pytest

from pygrambank.__main__ import main
from pygrambank.commands import new, issues, check_encoding, recode


@pytest.fixture
def args(api, wiki):
    return argparse.Namespace(repos=api, wiki_repos=wiki)


def test_recode_single_encoding():
    good_text = 'tèst'.encode('cp1252')
    bad_text = 'tèst'.encode('macroman')
    assert recode.decode_line(good_text, 'cp1252') == 'tèst'
    assert recode.decode_line(bad_text, 'cp1252') is None


def test_recode_mixed_encoding():
    good_text = b' -- '.join(['tèst'.encode('utf-8'), 'tèst'.encode('cp1252')])
    bad_text = b' -- '.join(['tèst'.encode('utf-8'), 'tèst'.encode('macroman')])
    assert recode.decode_line(good_text, 'mixed-cp1252') == 'tèst -- tèst'
    assert recode.decode_line(bad_text, 'mixed-cp1252') is None


def test_recode(tmp_path):
    good_single_enc = str(tmp_path / 'good_single_enc.txt')
    bad_single_enc = str(tmp_path / 'bad_single_enc.txt')
    good_mixed_enc = str(tmp_path / 'good_mixed_enc.txt')
    bad_mixed_enc = str(tmp_path / 'bad_mixed_enc.txt')

    with open(good_single_enc, 'wb') as f:
        f.write('tèst'.encode('cp1252'))
    with open(bad_single_enc, 'wb') as f:
        f.write('tèst'.encode('macroman'))
    with open(good_mixed_enc, 'wb') as f:
        f.write('tèst '.encode('cp1252'))
        f.write('tèst'.encode('utf-8'))
    with open(bad_mixed_enc, 'wb') as f:
        f.write('tèst '.encode('macroman'))
        f.write('tèst'.encode('utf-8'))

    main(['recode', '--encoding', 'cp1252', good_single_enc, bad_single_enc])
    main(['recode', '--encoding', 'mixed-cp1252', good_mixed_enc, bad_mixed_enc])

    with open(good_single_enc, 'rb') as f:
        assert f.read() == 'tèst\n'.encode('utf-8')
    with open(bad_single_enc, 'rb') as f:
        # file is unchanged
        assert f.read() == 'tèst'.encode('macroman')
    with open(good_mixed_enc, 'rb') as f:
        assert f.read() == 'tèst tèst\n'.encode('utf-8')
    with open(bad_mixed_enc, 'rb') as f:
        # file is unchanged
        assert f.read() == b' '.join([
            'tèst'.encode('macroman'),
            'tèst'.encode('utf-8')])


def test_updatefeatures(api, capsys):
    main(['--repos', str(api.repos),
          'updatefeatures', 'ABBR',
          '--wiki_repos', str(pathlib.Path(__file__).parent / 'grambank.wiki')])


def test_describe(api, capsys, tmp_path):
    main(['--repos', str(api.repos), 'describe', 'ABBR', '--columns'])
    out, _ = capsys.readouterr()
    assert 'abcd1234.tsv' in out
    p = tmp_path / 'ABBR_abcd1234.tsv'
    shutil.copy(str(api.sheets_dir.joinpath('ABBR_abcd1234.tsv')), str(p))
    main(['--repos', str(api.repos), 'describe', str(p), '--columns'])


def test_issues(args, capsys):
    args.id = None
    args.format = 'simple'
    issues.run(args)
    out, _ = capsys.readouterr()
    assert 'Comments' in out


def test_issues_detail(args, capsys):
    args.id = 223
    args.format = 'simple'
    issues.run(args)
    out, _ = capsys.readouterr()
    assert 'Ping @' in out


def test_new(args, tmp_path):
    args.out = sheet = tmp_path / 'sheet.tsv'
    new.run(args)
    assert sheet.exists()
    text = sheet.read_text(encoding='utf-8')
    assert all(s in text for s in ['GB021', 'Patron'])


def test_remove_empty(tmp_path):
    # contains empty columns and rows that need to be stripped
    content_a = (
        '\t\ta\t\tb\t\tc\t\t\t\n'
        '\t\t\t\t\t\t\t\t\t\n'
        '\t\t1\t\t2\t\t3\t\t\t\n'
        '\t\t\t\t\t\t\t\t\t\n'
        '\t\t\t\t\t\t\t\t\t\n')
    tsv_a = tmp_path / 'XX_aaaa1234.tsv'
    tsv_a.write_text(content_a, encoding="utf8")

    # contains rows that are too short
    content_b = 'a\tb\tc\n\n1\t2\nx\n'
    tsv_b = tmp_path / 'XX_bbbb1234.tsv'
    tsv_b.write_text(content_b, encoding="utf8")

    # contains missing column label that will raise an error
    content_c = 'a\tb\t\n1\t2\t3\nx\ty\tz\n'
    tsv_c = tmp_path / 'XX_cccc1234.tsv'
    tsv_c.write_text(content_c, encoding="utf8")

    main(['remove_empty', str(tsv_a), str(tsv_b)])
    assert tsv_a.read_text(encoding='utf8') == "a\tb\tc\n1\t2\t3\n"
    assert tsv_b.read_text(encoding='utf8') == 'a\tb\tc\n1\t2\t\nx\t\t\n'
    with pytest.raises(ValueError):
        main(['remove_empty', str(tsv_c)])
    # input file remains unchanged
    assert tsv_c.read_text(encoding='utf8') == 'a\tb\t\n1\t2\t3\nx\ty\tz\n'


def test_check(repos, capsys, tmp_path):
    main(['--repos', str(repos), 'check', '--verbose', '--report', str(tmp_path / 'report.tsv')])
    out, err = capsys.readouterr()
    assert all(word in out for word in ['WARNING', 'ERROR', 'Selecting', 'skipping'])
    main([
        '--repos', str(repos),
        'check',
        '--filename', str(repos / 'original_sheets' / 'ABBR_abcd1234.tsv')])


def test_sourcelookup(repos, capsys):
    main([
        '--repos', str(repos),
        'sourcelookup',
        str(pathlib.Path(__file__).parent / 'glottolog'),
        str(repos / 'original_sheets' / 'ABBR_abcd1234.tsv')])


def test_fix(repos):
    fname = repos / 'original_sheets' / 'ABBR_abcd1234.tsv'
    assert 'Author 2020' in fname.read_text(encoding='utf8')
    main([
        '--repos', str(repos),
        'fix',
        "lambda r: r.update(Source=(r.get('Source') or '').replace('2020', 'xyz')) or r",
        str(fname),
    ])
    text = fname.read_text(encoding='utf8')
    assert ('Author 2020' not in text) and ('Author xyz' in text)


def test_features(repos, capsys, mocker):
    mocker.patch('pygrambank.features.patrons', collections.defaultdict(lambda: 'HJH'))
    main([
        '--repos', str(repos), 'features',
        '--wiki_repos', str(pathlib.Path(__file__).parent / 'grambank.wiki'),
    ])
    out, err = capsys.readouterr()
    assert 'Patron' in out


def test_check_encoding_for_repl_chars():
    line = "�We don't want those dreaded ��� characters anywhere in our stuff!�"
    matches = list(check_encoding.find_replacement_chars(line))
    assert matches[0] == "�We don't want those "
    assert matches[1] == ' want those dreaded ��� characters anywhere'
    assert matches[2] == 'ywhere in our stuff!�'


def test_check_encoding_for_non_ascii():
    line = b"w\xe9 don't want non-utf-8 e\xf1\xe7\xf6ding anywhere in our st\xfcff"
    matches = list(check_encoding.find_non_ascii(line))
    assert matches[0] == b"w\xe9 don't want non-utf-"
    assert matches[1] == b"n't want non-utf-8 e\xf1\xe7\xf6ding anywhere in our"
    assert matches[2] == b'g anywhere in our st\xfcff'


def test_check_encoding_suggestions():
    line = b'wr\xf6ng enc\xf6ding'
    suggestions = check_encoding.suggest_encodings(line)
    assert suggestions['cp1252'] == 'wröng encöding'
    assert suggestions['macroman'] == 'wrˆng encˆding'


def test_check_encoding(tmp_path):
    good_utf8 = str(tmp_path / 'good_utf8.txt')
    corrupted_utf8 = str(tmp_path / 'corrupted-utf8.txt')
    not_utf8 = str(tmp_path / 'not-utf8.txt')
    mixed_encodings = str(tmp_path / 'mixed-encodings.txt')

    with open(good_utf8, 'wb') as f:
        f.write('Zum Gruße!\n'.encode('utf-8'))
    with open(corrupted_utf8, 'wb') as f:
        f.write('Zum Gu�e!\n'.encode('utf-8'))
    with open(not_utf8, 'wb') as f:
        f.write('Zum Gruße!\n'.encode('cp1252'))
    with open(mixed_encodings, 'wb') as f:
        f.write('Zum Gruße!\n'.encode('utf-8'))
        f.write('Zum Gruße!\n'.encode('cp1252'))

    main(['check_encoding', good_utf8, corrupted_utf8, not_utf8, mixed_encodings])



"""
Recode a file, i.e. change its encoding to UTF-8.
"""
from clldutils.clilib import PathType
import sys


def register(parser):
    parser.add_argument('path', nargs='+', type=PathType(type='file'))
    parser.add_argument(
        '--encoding',
        default='cp1252',
        choices=['cp1252', 'macroman', 'mixed-cp1252', 'mixed-macroman'],
    )


def _split_into_nonascii(raw_line):
    start = 0
    found_non_ascii = False
    for index, byte in enumerate(raw_line):
        # non-ascii
        if byte >= 0x80:
            found_non_ascii = True
        elif found_non_ascii:
            yield raw_line[start:index]
            found_non_ascii = False
            start = index
    if index > start:
        yield raw_line[start:]


def split_into_nonascii(raw_line):
    return list(_split_into_nonascii(raw_line))


def _decode_substring(raw_substring, enc):
    try:
        return raw_substring.decode('utf-8')
    except:  # noqa: E722
        try:
            return raw_substring.decode(enc)
        except:   # noqa: E722
            return None


def decode_mixed(raw_line, enc):
    substrings = split_into_nonascii(raw_line)
    unicode_strings = [_decode_substring(s, enc) for s in substrings]
    if None in unicode_strings:
        return None
    else:
        return ''.join(unicode_strings)


def decode_line(raw_line, enc):
    if enc.startswith('mixed-'):
        return decode_mixed(raw_line, enc[6:])
    else:
        try:
            return raw_line.decode(enc)
        except:  # noqa: E722
            return None


def run(args):
    for p in args.path:
        with open(p, 'rb') as f:
            raw_lines = [line.rstrip(b'\r\n') for line in f]

        unicode_lines = [decode_line(line, args.encoding) for line in raw_lines]
        found_invalid = False
        for index, (line, raw_line) in enumerate(zip(unicode_lines, raw_lines)):
            if line is None:
                found_invalid = True
                print(
                    '{}:{}: Could not decode line:'.format(p, index + 1),
                    repr(raw_line),
                    file=sys.stderr)

        if not found_invalid:
            with open(p, 'w', encoding='utf-8') as f:
                print('\n'.join(unicode_lines), file=f)



"""

"""
from csvw.dsv import UnicodeWriter, reader
from clldutils.clilib import PathType


def register(parser):
    parser.add_argument('sheet', type=PathType(type='file'))


def run(args):
    rows = list(reader(args.sheet, delimiter='\t'))
    with UnicodeWriter(args.sheet, delimiter='\t') as w:
        w.writerows(rows)



from pygrambank import bib


def test_author_extraction_from_bibkeys():
    assert list(bib.bibkey_authors('h:MacIntirePostman:stuff')) == ['MacIntire', 'Postman']
    assert list(bib.bibkey_authors('Meier2018')) == ['Meier']


def test_empty_author_string_yields_no_authors():
    author_string = ''
    authors = []
    assert list(bib.parse_authors(author_string)) == authors


def test_simple_first_name_plus_last_name():
    author_string = 'Thomas Mueller'
    authors = [{'firstname': 'Thomas', 'lastname': 'Mueller'}]
    assert list(bib.parse_authors(author_string)) == authors


def test_simple_last_name_plus_first_name():
    author_string = 'Mueller, Thomas'
    authors = [{'firstname': 'Thomas', 'lastname': 'Mueller'}]
    assert list(bib.parse_authors(author_string)) == authors


def test_multiple_authors():
    author_string = 'T. Mueller and H. Meier'
    authors = [
        {'firstname': 'T.', 'lastname': 'Mueller'},
        {'firstname': 'H.', 'lastname': 'Meier'},
    ]
    assert list(bib.parse_authors(author_string)) == authors


def test_von_part():
    author_string = 'Carl-Friedrich von Adelswegen'
    authors = [{'firstname': 'Carl-Friedrich', 'lastname': 'von Adelswegen'}]
    assert list(bib.parse_authors(author_string)) == authors

    author_string = 'von Adelswegen, Carl-Friedrich'
    authors = [{'firstname': 'Carl-Friedrich', 'lastname': 'von Adelswegen'}]
    assert list(bib.parse_authors(author_string)) == authors

    author_string = 'Adelswegen, Jr., Carl-Friedrich von'
    # is that really intended behaviour...?
    authors = [{
        'firstname': 'Carl-Friedrich von',
        'lastname': 'Adelswegen',
        'jr': 'Jr.'
    }]
    assert list(bib.parse_authors(author_string)) == authors



from pygrambank import features


def test_gb20(api, wiki, capsys):
    gb20 = features.GB20(api.repos / 'gb20.txt')
    api.repos.joinpath('For_coders').mkdir()
    gb20.listallwiki(wiki)


def test_gb20_update(api):
    def visit(f, api):
        f['xyz'] = 'abcdefg'
        return f

    api.visit_feature(visit)
    new = (api.repos / 'gb20.txt').read_text(encoding='utf-8')
    assert 'xyz' in new
    assert 'abcdefg' in new



"""
Removes empty columns and rows from a tsv file.
"""
from csvw import dsv
from clldutils.clilib import PathType
from pygrambank.sheet import Sheet
from itertools import chain, islice, repeat


def real_row_width(row):
    """Return width of a table row ignoring any empty cells at the end."""
    last_index = None
    for index, cell in enumerate(row):
        if cell.strip():
            last_index = index
    if last_index is None:
        return 0
    else:
        return last_index + 1


def real_table_height(table):
    """Return height of a table ignoring any empty rows at the end."""
    last_index = None
    for index, row in enumerate(table):
        if any(c.strip() for c in row):
            last_index = index
    if last_index is None:  # pragma: nocover
        return 0
    else:
        return last_index + 1


def padded_row(row, width):
    """Ensure size of a table row.

    If `row` is longer than `width`, truncate it.
    If `row` is shorter than `width`, pad it with empty strings.
    """
    return list(chain(
        islice(row, width),
        repeat('', max(0, width - len(row)))))


def register(parser):
    parser.add_argument('path', nargs='+', type=PathType(type='file'))


def run(args):
    for p in args.path:
        # use reader rather than iterrows so we operate on raw file rather than a
        # grambank-ifyed version.
        rows = list(Sheet(p)._reader())
        if not rows:  # pragma: nocover
            continue

        width = max(map(real_row_width, rows))
        height = real_table_height(rows)
        rows = [padded_row(row, width) for row in islice(rows, height)]

        cols_with_values = [
            x
            for x in range(width)
            if any(row[x] for row in rows)]
        unlabelled_cols = [
            x
            for x in range(width)
            if not rows[0][x] and x in cols_with_values]
        if unlabelled_cols:
            raise ValueError(
                '{}: Columns {} unlabelled. Fix manually.'.format(
                    p, ', '.join(map(str, unlabelled_cols))))

        new_rows = [
            [row[i] for i in range(width) if i in cols_with_values]
            for row in rows]
        new_rows = [row for row in new_rows if any(row)]

        with dsv.UnicodeWriter(p, delimiter='\t', encoding='utf8') as w:
            w.writerows(new_rows)



"""
Check original_sheets/ for correctness.
"""
from csvw.dsv import UnicodeWriter
from clldutils.clilib import PathType

from pygrambank.sheet import Sheet
from pygrambank.util import iterunique


def register(parser):
    parser.add_argument(
        '--filename',
        help="Path of a specific TSV file to check",
        default=None,
        type=PathType(type='file'),
    )
    parser.add_argument(
        '--report',
        help="Path of TSV file, to which results will be written.",
        default=None,
    )
    parser.add_argument(
        '--verbose',
        default=False,
        action='store_true',
    )


def run(args):
    report, counts = [], {}
    api = args.repos

    if args.filename:
        sheets = [Sheet(args.filename)]
    else:
        sheets = [(s, list(s.itervalues(api))) for s in api.iter_sheets()]
        sheets = (s[0] for s in iterunique(sheets, verbose=args.verbose))

    for sheet in sorted(sheets, key=lambda s: s.path):
        n = sheet.check(api, report=report)
        if (sheet.glottocode not in counts) or (n > counts[sheet.glottocode][0]):
            counts[sheet.glottocode] = (n, sheet.path.stem)

    selected = set(v[1] for v in counts.values())
    for row in report:
        row.insert(1, row[0] in selected)

    if report and args.report:
        with UnicodeWriter(args.report, delimiter='\t') as w:
            w.writerow(['sheet', 'selected', 'level', 'line', 'feature', 'message'])
            w.writerows(report)
        args.log.info('Report written to {0}'.format(args.report))
    args.log.error('Repos check found WARNINGs or ERRORs')



"""

"""
import textwrap
import collections

from clldutils.clilib import Table, add_format
from termcolor import colored


def register(parser):
    add_format(parser, 'simple')
    parser.add_argument('--id', help='Issue ID', type=int, default=None)


def run(args):
    issues = collections.OrderedDict([(i.id, i) for i in args.repos.issues])

    if args.id:
        issue = issues[args.id]
        print(colored('{0.created} {0.user}'.format(issue), 'green'))
        print(colored(issue.title, attrs={'bold', 'underline'}))
        print('')
        print(issue.body)
        for comment in issue.comments:
            print(colored('\n    {0.created} {0.user}'.format(comment), 'green'))
            for line in comment.stripped_body.split('\n'):
                print('    ' + line)
        return

    with Table(args, 'ID', 'Title', 'User', 'Comments') as t:
        for issue in issues.values():
            if issue.coding_problem:
                t.append([
                    issue.id,
                    textwrap.shorten(issue.title, 50, placeholder=' […]'),
                    issue.user,
                    len(issue.comments)])



"""
Applies fixes to sheets imported from the HG project. Rows to fix are identified by
- matching (Feature_ID, Value) against a list of known problems and
- checking whether the original "Autotranslated" comment is still there.
"""
import collections

from csvw.dsv import reader


class Fix:  # pragma: no cover
    def __init__(self, spec):
        self.spec = collections.defaultdict(list)
        for r in reader(spec, dicts=True):
            if r['find']:
                assert ':' in r['find'], r['find']
                self.spec[tuple(r['find'].split(':'))].append(r)

    def __call__(self, row):
        for k, v in row.items():
            if ('ontributed' in k) and ('atapoint' in k):
                if v.strip():
                    return True
        for key in [(row['Feature_ID'], row['Value']), (row['Feature_ID'], '*')]:
            if key in self.spec:
                # We found a matching datapoint ...
                for spec in self.spec[key]:
                    if spec['Mapping'] in row['Comment'] and ('MM' not in row['Comment']):
                        # ... with a matching "Autotranslated" comment.
                        fid, val = spec['replace'].replace('GB86', 'GB086').split(':')
                        assert fid == row['Feature_ID'], spec
                        row['Value'] = val
                        if not val.strip():
                            row['Comment'] = ''
                            row['Source'] = ''
                        break
        return True


def run(args):  # pragma: no cover
    fix = Fix(args.repos.repos / 'HG_import' / 'HG_GB_mapping_JV_HS_revise.csv')
    for sheet in args.repos.iter_sheets():
        if sheet.path.stem.startswith('CB-'):
            sheet.visit(row_visitor=fix)



"""
Create a new, empty sheet
"""
import pathlib
import collections

from csvw.dsv import UnicodeWriter

from pygrambank.cli_util import add_wiki_repos

FNAME = 'For_coders/Grambank_most_updated_sheet.tsv'


def register(parser):
    parser.add_argument(
        '--out',
        help='Path for output sheet',
        default=FNAME)
    add_wiki_repos(parser)


def normalised_summary(f):
    summary = f.wiki['Summary']
    summary = summary.replace('\n', ' ')
    summary = summary.replace('\t', ' ')
    return summary


def run(args):
    if args.out == FNAME:  # pragma: no cover
        name = args.repos.path(args.out)
    else:
        name = pathlib.Path(args.out)

    rows = collections.OrderedDict([
        ('Feature_ID', lambda f: f.id),
        ('Feature', lambda f: f.wiki['title']),
        ('Possible Values', lambda f: f['Possible Values']),
        ('Language_ID', lambda f: ''),
        ('Value', lambda f: ''),
        ('Source', lambda f: ''),
        ('Comment', lambda f: ''),
        ('Contributed_Datapoints', lambda f: ''),
        ('Clarifying comments', normalised_summary),
        ('Relevant unit(s)', lambda f: f['Relevant unit(s)']),
        ('Function', lambda f: f['Function']),
        ('Form', lambda f: f['Form']),
        ('Patron', lambda f: f.wiki['Patron']),
    ])

    with UnicodeWriter(name, delimiter='\t') as w:
        w.writerow(rows.keys())
        for feature in args.repos.features.values():
            w.writerow([v(feature) for v in rows.values()])



import re

import dateutil.parser
from markdown import markdown
from bs4 import BeautifulSoup as bs

# Redirect links:
# wiki:
# https://github.com/glottobank/Grambank/wiki/Responsibilities-for-parties-involved-in-GramBank
# issues:
# https://github.com/glottobank/Grambank/issues/114
# FIXME:
# - inline images
# - render threads as html

__all__ = ['Issue']


class Base:
    def __init__(self, d):
        self.d = d
        self.body = self.d['body'].strip()

    @property
    def stripped_body(self):
        return '\n'.join([
            line for line in self.body.split('\n') if not line.startswith('> ')]).strip()

    @property
    def user(self):
        return self.d['user']['login']

    @property
    def created(self):
        return dateutil.parser.isoparse(self.d['created_at'])

    @property
    def html(self):
        return markdown(self.d['body'])

    def iter_urls(self):
        for m in re.finditer(r'(?P<url>http(s)?://[^\s)]+)', self.stripped_body):
            yield m.group('url')

    def iter_issue_links(self):
        for m in re.finditer('(?P<url>#[0-9]+)', self.stripped_body):
            yield m.group('url')

    def iter_links(self):
        for link in bs(self.html, 'html5lib').find_all('a'):
            yield link['href'], link.text

    def iter_images(self):
        for link in bs(self.html, 'html5lib').find_all('img'):
            yield link['src'], link.get('alt')


class Comment(Base):
    pass


class Issue(Base):
    def __init__(self, d, comments):
        Base.__init__(self, d)
        self._comments = comments
        assert self.d['comments'] == len(self._comments), '{}: {} vs {}'.format(
            self.d['number'], self.d['comments'], len(self._comments))
        self.id = self.d['number']

    @property
    def title(self):
        return self.d['title']

    @property
    def coding_problem(self):
        for label in self.d['labels']:
            if label['name'] == 'coding problem':
                return True
        return False  # pragma: no cover

    @property
    def comments(self):
        return [Comment(comment) for comment in self._comments]



"""

"""
import itertools
import subprocess
import collections

from csvw.dsv import reader, UnicodeWriter
from clldutils.misc import nfilter

from .check_conflicts import check

# flake8: noqa

CODERS = collections.OrderedDict([
    ('SydneyRey', 'SR'),
    ('Michael90EVAMPI', 'MM'),
    ('Michael', 'MM'),
    ('Jill', 'JSA'),
    ('vickygruner', 'VG'),
    ('Hedvig', 'HS'),
])


def get_coder(p):
    log = subprocess.check_output(['git', 'log', str(p)]).decode('utf8')
    committers = []
    for line in log.split('\n'):
        if line.startswith('Author:'):
            committers.append(line.replace('Author:', '').strip().split()[0])
    for name, initials in CODERS.items():
        if name in committers:
            return initials


def merged_rows(rows, active):
    if all(r['Conflict'].lower().startswith('true') for r in rows):
        for row in rows:
            if row['Select'].lower().strip() == 'true':
                return row
        assert rows[0]['Feature_ID'] not in active, str(rows)
        return None
    elif all(r['Conflict'].lower().strip() == 'false' for r in rows):
        row = rows[0]
        for k in ['Sheet', 'Comment', 'Source']:
            row[k] = [row[k]]
        for r in rows[1:]:
            for k in ['Sheet', 'Comment', 'Source']:
                row[k].append(r[k])
        for k, sep in [('Sheet', ' '), ('Comment', '. '), ('Source', '; ')]:
            row[k] = sep.join(nfilter(row[k]))
        return row
    raise ValueError(rows)


def rows_and_sourcesheets(sheet, active):
    allrows, sourcesheets = [], []

    def clean_row(row):
        sourcesheets.extend(row['Sheet'].split())
        coders = '-'.join(s.split('_')[0] for s in row['Sheet'].split())
        coders = coders.split('-')
        del row['Sheet']
        if row.get('Contributed_Datapoint'):
            row['Contributed_Datapoint'] += ' ' + ' '.join(coders)
        else:
            row['Contributed_Datapoint'] = ' '.join(coders)
        allrows.append(row)

    #fids = collections.Counter([r['Feature_ID'] for r in reader(sheet, dicts=True, delimiter='\t')])
    #print(len(fids), sum(fids.values()))

    for fid, rows in itertools.groupby(
        sorted(reader(sheet, dicts=True, delimiter='\t'), key=lambda r: r['Feature_ID']),
        lambda r: r['Feature_ID'],
    ):
        rows = list(rows)
        if len(rows) == 1:
            clean_row(rows[0])
        else:
            row = merged_rows(rows, active)
            if row:
                clean_row(row)
    return allrows, set(sourcesheets)


def write(p, rows, features):
    rows = collections.OrderedDict([(r['Feature_ID'], r) for r in rows])
    cols = 'Feature_ID Value Source Contributed_Datapoint Comment'.split()
    with UnicodeWriter(p, delimiter='\t') as writer:
        writer.writerow(cols)

        for feature in features.values():
            row = rows.pop(feature.id, None)
            if not row:
                row = {k: '' for k in cols}
                row['Feature_ID'] = feature.id
            writer.writerow([row.get(col, '') for col in cols])
        for row in rows.values():
            writer.writerow([row.get(col, '') for col in cols])


def run(args):
    active = list(args.repos.features)

    for sheet in sorted(args.repos.path('conflicts').glob('*.tsv'), key=lambda p: p.name):
        #if sheet.stem != 'brah1256':
        #    continue
        ok, nc = check(sheet)
        if ok and nc:
            print(sheet.stem)
            rows, sources = rows_and_sourcesheets(sheet, active)
            coder = get_coder(sheet)
            if sheet.stem == 'sout2989':
                coder = 'JE-HS'
            assert coder, str(sheet)
            write(args.repos.path('original_sheets', '{}_{}.tsv'.format(coder, sheet.stem)), rows, args.repos.features)
            for src in sources:
                if src.split('_')[0] != coder:
                    p = args.repos.path('original_sheets', src.split('.')[0] + '.tsv')
                    if p.exists():
                        p.unlink()
                    else:
                        args.log.warning('Unknown source sheet: {}'.format(p))



"""
"""
import itertools

from csvw.dsv import reader
from clldutils.clilib import PathType


def register(parser):
    parser.add_argument('sheet', type=PathType(type='file'))


classes = {
    'correct': 'this is the correct value for this datapoint',
    'typo': 'there is no apparent reason, given the source(s) and the comment the inaccurate '
            'coding appears to be a typo',
    'overcautious': '',
    'overconfident': '',
    'new source was consulted in other coding': '',
    'feature interpretation': "the different coders have interpreted the feature differently "
                              "(some perhaps coded outside of the project and weren't in the "
                              "loop with the specifics)",
    'source interpretation': "the different coders have interpreted the source(s) differently",
    'sources disagree': '',
    'missed': "some coders have missed a section in a source",
    'less thorough coding': "the other coder(s) were particularly thorough and spotted "
                            "something that could easily have been missed or contacted expert(s)",
    'hard to code': "the language phenomena is difficult to neatly fit into the questionnaire "
                    "and took considerable discussion to resolve",
    'unclear': "it is not clear why there is a conflict",
}


def norm(r, k):
    return r[k].strip().lower()


def check(sheet, log=None):
    ok = True
    conflicts = 0

    def error(msg):
        if log:
            log.error(msg)
        return False

    for fid, rows in itertools.groupby(
            sorted(reader(sheet, dicts=True, delimiter='\t'), key=lambda r: r['Feature_ID']),
            lambda r: r['Feature_ID'],
    ):
        rows = [(r, norm(r, 'Conflict'), norm(r, 'Select'), norm(r, 'Classification of conflict'))
                for r in rows]
        if all(r[1] == 'true' for r in rows):
            selected_value = set(r[0]['Value'] for r in rows if r[2] == 'true')
            if len(selected_value) != 1:
                ok = error('Different values selected for {}'.format(fid))

            # For every set of same features where there is "True" for conflict, at least one should
            # have "True" in the col "Select"
            if not any(r[2] == 'true' for r in rows):
                ok = error('No selected row for {}'.format(fid))

            for row, _, select, cls in rows:
                # For every feature where there is "True" for conflict and not "True" for "Select",
                # there should be something in the col "classification of conflict" that comes
                # from this specific list
                if select != 'true':
                    if cls not in classes:
                        ok = error('Invalid classification for {}: {}'.format(fid, cls))

                # Every feature which is True for select should also have "Correct" in
                # "classification of conflict"
                if select == 'true':
                    if cls != 'correct':
                        ok = error('Selected row not classified as correct for {}'.format(fid))
                    if not row['Contributed_Datapoint']:
                        ok = error('Selected row has no code specified in Contributed_Datapoint')

                # Every feature where there is "True" for conflict and "Correct" in
                # "classification of conflict" but nothing in "Select" should have the same Value
                # as whichever row has "Select" =" True in that set of features for that language
                # (this only comes into effect if there is a comparison
                # of more than 2 sheets and one row happens to be better Comment or Source-wise).
                if cls == 'correct' and not select:
                    if row['Value'] not in selected_value:
                        ok = error('Row classified as correct, but value is different from '
                                   'selected value for {}'.format(fid))

            conflicts += 1

    return ok, conflicts


def run(args):
    ok, conflicts = check(args.sheet, args.log)
    if ok:
        args.log.info('{} conflicts'.format(conflicts))
    else:
        args.log.warning('There were errors in the {} conflicts'.format(conflicts))
